# SPDX-FileCopyrightText: 2019-2026 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

import base64
from gettext import gettext as _
import json
import logging
import time
from urllib.parse import quote
from urllib.parse import unquote

from bs4 import BeautifulSoup
import requests

from komikku.consts import USER_AGENT
from komikku.servers import Server
from komikku.utils import get_buffer_mime_type

logger = logging.getLogger(__name__)


def b36encode(number, alphabet='0123456789abcdefghijklmnopqrstuvwxyz'):
    base36 = ''
    sign = ''

    if number < 0:
        sign = '-'
        number = -number

    if 0 <= number < len(alphabet):
        return sign + alphabet[number]

    while number != 0:
        number, i = divmod(number, len(alphabet))
        base36 = alphabet[i] + base36

    return sign + base36


class Bdplus(Server):
    id = 'bdplus'
    name = 'BDplus'
    lang = 'fr'

    base_url = 'https://bdplus.cc'
    logo_url = base_url + '/favicon.ico'
    search_url = base_url + '/catalogue.html'
    manga_url = base_url + '/serie_static/{0}.html'
    chapter_url = base_url + '/lecture.html?b={0}'
    chapter_pages_url = base_url + '/_converted/{0}/{1}/pages.json'
    image_url = base_url + '/_converted/{0}/{1}/{2}'

    filters = [
        {
            'key': 'genre',
            'type': 'select',
            'name': _('Genre'),
            'description': _('Filter by Genre'),
            'value_type': 'single',
            'default': 'Tous',
            'options': [
                {'key': 'Tous', 'name': 'Tous'},
                {'key': 'Science-Fiction', 'name': 'Science-Fiction'},
                {'key': 'Fantasy', 'name': 'Fantasy'},
                {'key': 'Aventure', 'name': 'Aventure'},
                {'key': 'Humour', 'name': 'Humour'},
                {'key': 'Historique', 'name': 'Historique'},
                {'key': 'Thriller / Policier', 'name': 'Thriller / Policier'},
                {'key': 'Horreur / Mystique', 'name': 'Horreur / Mystique'},
                {'key': 'Western', 'name': 'Western'},
                {'key': 'Sport', 'name': 'Sport'},
                {'key': 'Autre', 'name': 'Autre'},
            ],
        },
        {
            'key': 'epoque',
            'type': 'select',
            'name': 'Époque',
            'description': 'Filtrer par époque',
            'value_type': 'single',
            'default': 'Toutes',
            'options': [
                {'key': 'Toutes', 'name': 'Toutes'},
                {'key': 'Classiques', 'name': 'Classiques (<1980)'},
                {'key': '1980s', 'name': 'Années 80-90'},
                {'key': '2000s', 'name': 'Années 2000'},
                {'key': '2010s', 'name': 'Années 2010'},
                {'key': 'Recentes', 'name': 'Récentes (2020+)'},
            ],
        },
    ]

    manga_list = []

    def __init__(self):
        if self.session is None:
            self.session = requests.Session()
            self.session.headers.update({'User-Agent': USER_AGENT})

    def get_manga_data(self, initial_data):
        """
        Returns manga data by scraping manga HTML page content

        Initial data should contain at least manga's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Manga slug is missing in initial data'

        r = self.session_get(self.manga_url.format(initial_data['slug']))
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = initial_data.copy()
        data.update({
            'authors': [],
            'scanlators': [],  # Not available
            'genres': [],
            'status': None,  # Not available
            'synopsis': None,
            'chapters': [],
            'server_id': self.id,
            'cover': None,
        })

        data['name'] = soup.select_one('h1.serie-main-title').text.strip()
        if element := soup.select_one('img.serie-cover'):
            data['cover'] = f'{self.base_url}{element.get("src")}'

        # Details
        if element := soup.select_one('.meta-authors'):
            for author in element.text.split(','):
                author = author.strip()
                if author not in data['authors']:
                    data['authors'].append(author)

        for element in soup.select('.genre-pill'):
            data['genres'].append(element.text.strip())

        if element := soup.select_one('.synopsis-text'):
            synopsis = [element.text.strip(), ]
            if element := soup.select_one('#synopsis-more'):
                soup_synopsis = BeautifulSoup(element.get('data-content'), 'lxml')
                for element in soup_synopsis.select('p'):
                    synopsis.append(element.text.strip())
            data['synopsis'] = '\n\n'.join(synopsis)

        # Chapters
        for element in soup.select('#tomes-grid a.card'):
            data['chapters'].append({
                'slug': element.get('data-path').split('/')[-1],
                'title': element.select_one('.card-title').text.strip(),
            })

        return data

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data
        """

        # Compute encoded path
        # See `encodePathObfuscated` func of `/assets/js/serie_static_secure.min.xxxxxxxxxx.js` script
        path = f'_converted/{manga_slug}/{chapter_slug}'
        t = 3
        a = b36encode(int(time.time() * 1000)).lower()
        n = base64.b64encode(unquote(quote(f'{path}/{a}')).encode()).decode()
        t %= len(n)
        encoded_path = n[-t:] + n[:-t]

        r = self.session_get(
            self.chapter_pages_url.format(manga_slug, chapter_slug),
            headers={
                'Referer': self.chapter_url.format(encoded_path),
            }
        )
        if r.status_code != 200:
            return None

        data = {
            'pages': [],
        }
        for page in r.json()['pages']:
            data['pages'].append({
                'slug': None,
                'image': page['url'],
            })

        return data

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        r = self.session_get(
            self.image_url.format(manga_slug, chapter_slug, page['image']),
            headers={
                'Referer': f'{self.base_url}/',
            }
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return {
            'buffer': r.content,
            'mime_type': mime_type,
            'name': page['image'].split('/')[-1],
        }

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return self.manga_url.format(slug)

    def get_latest_updates(self, genre=None, epoque=None):
        """
        Returns latest updates
        """
        r = self.session_get(self.base_url)
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        results = []
        for element in soup.select('#nouveautes-carousel a.nouveaute-card'):
            results.append({
                'slug': element.get('href').split('/')[-1].replace('.html', ''),
                'name': element.select_one('.card-title').text.strip(),
                'cover': self.base_url + element.select_one('img').get('src'),
            })

        return results

    def search(self, term=None, genre=None, epoque=None):
        if not self.manga_list:
            # Retrieve manga list if not already done
            r = self.session_get(self.search_url)
            if r.status_code != 200:
                return None

            soup = BeautifulSoup(r.text, 'lxml')

            if script_element := soup.select_one('script#catalogue-data'):
                if data := script_element.string:
                    self.manga_list = json.loads(data)

        if not self.manga_list:
            return None

        results = []
        for serie in self.manga_list:
            if term and term.lower() not in serie[1].lower():
                continue

            if genre and genre != 'Tous' and genre != serie[3]:
                continue

            if epoque and epoque != 'Toutes':
                if epoque == 'Classiques' and serie[6] >= 1980:
                    continue
                if epoque == '1980s' and (serie[6] < 1980 or serie[6] > 1999):
                    continue
                if epoque == '2000s' and (serie[6] < 2000 or serie[6] > 2009):
                    continue
                if epoque == '2010s' and (serie[6] < 2010 or serie[6] > 2019):
                    continue
                if epoque == 'Recentes' and serie[6] < 2020:
                    continue

            results.append({
                'slug': serie[0],
                'name': serie[1],
                'cover': self.base_url + serie[2],
            })

        return results



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>
# Author: Arthur Williams (TAAPArthur)

import urllib.parse
import re

from bs4 import BeautifulSoup
import piexif
from PIL import Image

from komikku.consts import USER_AGENT
from komikku.servers import Server
from komikku.servers.utils import convert_date_string
from komikku.servers.utils import do_login
from komikku.utils import is_number

# Improved from https://github.com/manga-py/manga-py
# This server uses a region restriction, a VPN can be required


class Vizmanga(Server):
    id = 'vizmanga'
    name = 'VizManga'
    lang = 'en'
    locale = 'enUS'
    has_login = True
    status = 'disabled'  # commercial platforms are no supported

    base_url = 'https://www.viz.com'
    logo_url = base_url + '/favicon/favicon-32x32.png?v=47MPqANpyj'
    login_url = base_url + '/account/try_login'
    refresh_login_url = base_url + '/account/refresh_login_links'
    api_series_url = base_url + '/shonenjump'
    api_series_search_url = base_url + '/search?search={}'
    api_chapters_url = base_url + '/shonenjump/chapters/{}'
    api_chapter_data_url = base_url + '/manga/get_manga_url?device_id=3&manga_id={}&page={}'
    api_chapter_url = base_url + '/shonenjump/{}-chapter-{}/chapter/{}'

    chapter_regex = re.compile(r'/shonenjump/(.*)-chapter-([\d\-]*)/chapter/(\d*)')
    page_regex = re.compile(r'var pages\s*=\s*(\d*);')

    headers = {
        'Referer': 'https://www.viz.com',
        'Accept': 'text/html,application/xhtml+xml,application/xml;q=1.0,image/webp,image/apng,*/*;q=1.0',
        'User-Agent': USER_AGENT,
    }

    def __init__(self, username=None, password=None):
        if username and password:
            self.do_login(username, password)

    @do_login
    def get_manga_data(self, initial_data):
        """
        Returns manga data from API

        Initial data should contain at least manga's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Manga slug is missing in initial data'

        self.refresh_login()
        r = self.session_get(self.api_chapters_url.format(initial_data['slug']))
        soup = BeautifulSoup(r.text, 'lxml')

        authors = []
        synopsis = ''
        manga_info = soup.find('section', {'id': 'series-intro'})
        if manga_info:
            author_element = manga_info.find('span', {'class': 'disp-bl--bm mar-b-md'})
            authors = author_element.getText().split(',') if author_element else ''
            synposis_element = manga_info.find('h4')
            synopsis = synposis_element.getText().strip() if synposis_element else ''

        ongoing = soup.find('div', {'class': 'section_future_chapter'})
        status = 'ongoing' if ongoing else 'complete'

        data = initial_data.copy()
        data.update(dict(
            authors=authors,
            scanlators=[],
            genres=[],
            status=status,
            chapters=[],
            synopsis=synopsis,
            server_id=self.id,
            cover=initial_data['url'],
            url=initial_data['url'],
        ))

        # Chapters
        chapters = soup.findAll('a', {'class': 'o_chapter-container'})
        if soup.select_one('#chpt_rows').get('aria-sort') == 'descending':
            chapters.reverse()
        slugs = set()
        for chapter in chapters:
            raw_url_maybe = chapter['data-target-url']
            match = self.chapter_regex.search(raw_url_maybe)
            series_name = match.group(1)
            chapter_number = match.group(2)
            chapter_slug = match.group(3)
            chapter_date = None
            # There could be duplicate elements with the same chapter slug; they refer to the same chapter so skip them
            if chapter_slug in slugs:
                continue
            slugs.add(chapter_slug)
            try:
                chapter_date_str = chapter.find('td', {'class': 'pad-y-0 pad-r-0 pad-r-rg--sm'}).getText()
                chapter_date = convert_date_string(chapter_date_str, '%B %d, %Y')
            except AttributeError:
                pass
            # There seems to be a title field in the metadata... it doesn't seem useful nor unique
            # eg 'My Hero Academia: Vigilantes Chapter 1.0'
            # so we'll just use the chapter number as the title
            title = chapter_number

            data['chapters'].append(dict(
                slug=chapter_slug,
                title=f'Ch. {title}',
                url=self.api_chapter_url.format(series_name, chapter_number, chapter_slug),
                num=chapter_number if is_number(chapter_number) else None,
                date=chapter_date,
            ))

        return data

    @do_login
    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data

        Currently, only pages are expected.
        """

        if not self.refresh_login():
            return None
        match = self.page_regex.search(self.session_get(chapter_url).text)
        num_pages = int(match.group(1)) + 1

        pages = []
        for i in range(num_pages):
            url = self.api_chapter_data_url.format(chapter_slug, i)
            pages.append({'slug': None, 'image': url})
        return {'pages': pages}

    def _get_chapter_url(self, chapter_slug, i):
        url = self.api_chapter_data_url.format(chapter_slug, i)

        r = self.session_get(url)
        return r.text.strip()

    @do_login
    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        r = self.session_get(page['image'])
        real_img_url = r.text.strip()

        r = self.session_get(real_img_url, stream=True)
        if r.status_code != 200:
            return None

        orig = Image.open(r.raw)  # type: Image.Image
        solution = Vizmanga.solve_image(orig)

        buffer = solution
        mime_type = None

        return dict(
            buffer=buffer,
            mime_type=mime_type,
            name=real_img_url.split('/')[-1].split('?')[0],
        )

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return self.api_chapters_url.format(slug)

    @do_login
    def get_most_populars(self):
        """
        Returns full list of manga sorted by rank
        """
        r = self.session_get(self.api_series_url)
        return self._parse_page(r)

    def _parse_page(self, r, classes='o_chapters-link'):
        """
        Returns full list of manga sorted by rank
        """

        soup = BeautifulSoup(r.text, 'lxml')
        divs = soup.findAll('a', {'class': f'disp-bl color-white {classes}'})
        result = []
        for div in divs:
            slug = div['href'].split('/')[-1]
            name = next(iter(div.find_all('div', {'class', re.compile('pad-x-rg.*type-sm type-rg--sm type-md--lg type-center.*')}))).getText().strip()
            cover_url = div.find('img')['data-original']
            result.append({
                'slug': slug,
                'name': name,
                'url': cover_url,
            })

        return result

    def get_token(self):
        auth_token = self.session_get(self.refresh_login_url)
        token = re.search(r'AUTH_TOKEN\s*=\s*"(.+?)"', auth_token.text)
        return token.group(1)

    def refresh_login(self):
        r = self.session_get(self.refresh_login_url)
        soup = BeautifulSoup(r.text, 'lxml')
        return bool(soup.select('.o_profile-link'))

    def login(self, username, password):
        """
        Log in and initializes API
        """
        if not username or not password:
            return False
        token = self.get_token()

        r = self.session_post(
            self.login_url,
            data={
                'login': username,
                'pass': password,
                'rem_user': 1,
                'authenticity_token': token,
            })
        self.save_session()
        return r.status_code == 200

    @do_login
    def search(self, term):
        term_lower = urllib.parse.quote_plus(term.lower())
        r = self.session_get(self.api_series_search_url.format(term_lower))
        return list(self._parse_page(r, classes='o_property-link'))

    @staticmethod
    def solve_image(orig: Image) -> Image.Image:
        exif = piexif.load(orig.info['exif'])

        new_size = (orig.size[0] - 90, orig.size[1] - 140)
        ref = Image.new(orig.mode, new_size)  # type: Image.Image
        ref.paste(orig)

        _key = 42016
        key = [int(i, 16) for i in exif['Exif'][_key].decode('ascii').split(':')]
        width, height = exif['0th'][256], exif['0th'][257]

        small_width = int(width / 10)
        small_height = int(height / 15)

        Vizmanga.paste(ref, orig, (
            0, small_height + 10,
            small_width, height - 2 * small_height,
        ), (
            0, small_height,
            small_width, height - 2 * small_height,
        ))

        Vizmanga.paste(ref, orig, (
            0, 14 * (small_height + 10),
            width, orig.height - 14 * (small_height + 10),
        ), (
            0, 14 * small_height,
            width, orig.height - 14 * (small_height + 10),
        ))

        Vizmanga.paste(ref, orig, (
            9 * (small_width + 10), small_height + 10,
            small_width + (width - 10 * small_width), height - 2 * small_height,
        ), (
            9 * small_width, small_height,
            small_width + (width - 10 * small_width), height - 2 * small_height,
        ))

        for i, j in enumerate(key):
            Vizmanga.paste(ref, orig, (
                (i % 8 + 1) * (small_width + 10), (int(i / 8) + 1) * (small_height + 10),
                small_width, small_height,
            ), (
                (j % 8 + 1) * small_width, (int(j / 8) + 1) * small_height,
                small_width, small_height,
            ))

        return ref

    @staticmethod
    def paste(ref: Image.Image, orig: Image.Image, orig_box, ref_box):
        ref.paste(orig.crop((
            int(orig_box[0]), int(orig_box[1]),
            int(orig_box[0] + orig_box[2]), int(orig_box[1] + orig_box[3]),
        )), (
            int(ref_box[0]), int(ref_box[1]),
            int(ref_box[0] + ref_box[2]), int(ref_box[1] + ref_box[3]),
        ))



# SPDX-FileCopyrightText: 2019-2026 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

import json
import urllib.parse

import requests

from komikku.servers.exceptions import ServerException

# https://translate.google.com/translate_a/l?client=t&alpha=true
LANGUAGES = {'ab': 'Abkhaz', 'ace': 'Acehnese', 'ach': 'Acholi', 'aa': 'Afar', 'af': 'Afrikaans', 'sq': 'Albanian', 'alz': 'Alur', 'am': 'Amharic', 'ar': 'Arabic', 'hy': 'Armenian', 'as': 'Assamese', 'av': 'Avar', 'awa': 'Awadhi', 'ay': 'Aymara', 'az': 'Azerbaijani', 'ban': 'Balinese', 'bal': 'Baluchi', 'bm': 'Bambara', 'bci': 'Baoulé', 'ba': 'Bashkir', 'eu': 'Basque', 'btx': 'Batak Karo', 'bts': 'Batak Simalungun', 'bbc': 'Batak Toba', 'be': 'Belarusian', 'bem': 'Bemba', 'bn': 'Bengali', 'bew': 'Betawi', 'bho': 'Bhojpuri', 'bik': 'Bikol', 'bs': 'Bosnian', 'br': 'Breton', 'bg': 'Bulgarian', 'bua': 'Buryat', 'yue': 'Cantonese', 'ca': 'Catalan', 'ceb': 'Cebuano', 'ch': 'Chamorro', 'ce': 'Chechen', 'ny': 'Chichewa', 'zh-CN': 'Chinese (Simplified)', 'zh-TW': 'Chinese (Traditional)', 'chk': 'Chuukese', 'cv': 'Chuvash', 'co': 'Corsican', 'crh': 'Crimean Tatar (Cyrillic)', 'crh-Latn': 'Crimean Tatar (Latin)', 'hr': 'Croatian', 'cs': 'Czech', 'da': 'Danish', 'fa-AF': 'Dari', 'dv': 'Dhivehi', 'din': 'Dinka', 'doi': 'Dogri', 'dov': 'Dombe', 'nl': 'Dutch', 'dyu': 'Dyula', 'dz': 'Dzongkha', 'en': 'English', 'eo': 'Esperanto', 'et': 'Estonian', 'ee': 'Ewe', 'fo': 'Faroese', 'fj': 'Fijian', 'tl': 'Filipino', 'fi': 'Finnish', 'fon': 'Fon', 'fr': 'French', 'fr-CA': 'French (Canada)', 'fy': 'Frisian', 'fur': 'Friulian', 'ff': 'Fulani', 'gaa': 'Ga', 'gl': 'Galician', 'ka': 'Georgian', 'de': 'German', 'el': 'Greek', 'gn': 'Guarani', 'gu': 'Gujarati', 'ht': 'Haitian Creole', 'cnh': 'Hakha Chin', 'ha': 'Hausa', 'haw': 'Hawaiian', 'iw': 'Hebrew', 'hil': 'Hiligaynon', 'hi': 'Hindi', 'hmn': 'Hmong', 'hu': 'Hungarian', 'hrx': 'Hunsrik', 'iba': 'Iban', 'is': 'Icelandic', 'ig': 'Igbo', 'ilo': 'Ilocano', 'id': 'Indonesian', 'iu-Latn': 'Inuktut (Latin)', 'iu': 'Inuktut (Syllabics)', 'ga': 'Irish', 'it': 'Italian', 'jam': 'Jamaican Patois', 'ja': 'Japanese', 'jw': 'Javanese', 'kac': 'Jingpo', 'kl': 'Kalaallisut', 'kn': 'Kannada', 'kr': 'Kanuri', 'pam': 'Kapampangan', 'kk': 'Kazakh', 'kha': 'Khasi', 'km': 'Khmer', 'cgg': 'Kiga', 'kg': 'Kikongo', 'rw': 'Kinyarwanda', 'ktu': 'Kituba', 'trp': 'Kokborok', 'kv': 'Komi', 'gom': 'Konkani', 'ko': 'Korean', 'kri': 'Krio', 'ku': 'Kurdish (Kurmanji)', 'ckb': 'Kurdish (Sorani)', 'ky': 'Kyrgyz', 'lo': 'Lao', 'ltg': 'Latgalian', 'la': 'Latin', 'lv': 'Latvian', 'lij': 'Ligurian', 'li': 'Limburgish', 'ln': 'Lingala', 'lt': 'Lithuanian', 'lmo': 'Lombard', 'lg': 'Luganda', 'luo': 'Luo', 'lb': 'Luxembourgish', 'mk': 'Macedonian', 'mad': 'Madurese', 'mai': 'Maithili', 'mak': 'Makassar', 'mg': 'Malagasy', 'ms': 'Malay', 'ms-Arab': 'Malay (Jawi)', 'ml': 'Malayalam', 'mt': 'Maltese', 'mam': 'Mam', 'gv': 'Manx', 'mi': 'Maori', 'mr': 'Marathi', 'mh': 'Marshallese', 'mwr': 'Marwadi', 'mfe': 'Mauritian Creole', 'chm': 'Meadow Mari', 'mni-Mtei': 'Meiteilon (Manipuri)', 'min': 'Minang', 'lus': 'Mizo', 'mn': 'Mongolian', 'my': 'Myanmar (Burmese)', 'nhe': 'Nahuatl (Eastern Huasteca)', 'ndc-ZW': 'Ndau', 'nr': 'Ndebele (South)', 'new': 'Nepalbhasa (Newari)', 'ne': 'Nepali', 'bm-Nkoo': 'NKo', 'no': 'Norwegian', 'nus': 'Nuer', 'oc': 'Occitan', 'or': 'Odia (Oriya)', 'om': 'Oromo', 'os': 'Ossetian', 'pag': 'Pangasinan', 'pap': 'Papiamento', 'ps': 'Pashto', 'fa': 'Persian', 'pl': 'Polish', 'pt': 'Portuguese (Brazil)', 'pt-PT': 'Portuguese (Portugal)', 'pa': 'Punjabi (Gurmukhi)', 'pa-Arab': 'Punjabi (Shahmukhi)', 'qu': 'Quechua', 'kek': 'Qʼeqchiʼ', 'rom': 'Romani', 'ro': 'Romanian', 'rn': 'Rundi', 'ru': 'Russian', 'se': 'Sami (North)', 'sm': 'Samoan', 'sg': 'Sango', 'sa': 'Sanskrit', 'sat-Latn': 'Santali (Latin)', 'sat': 'Santali (Ol Chiki)', 'gd': 'Scots Gaelic', 'nso': 'Sepedi', 'sr': 'Serbian', 'st': 'Sesotho', 'crs': 'Seychellois Creole', 'shn': 'Shan', 'sn': 'Shona', 'scn': 'Sicilian', 'szl': 'Silesian', 'sd': 'Sindhi', 'si': 'Sinhala', 'sk': 'Slovak', 'sl': 'Slovenian', 'so': 'Somali', 'es': 'Spanish', 'su': 'Sundanese', 'sus': 'Susu', 'sw': 'Swahili', 'ss': 'Swati', 'sv': 'Swedish', 'ty': 'Tahitian', 'tg': 'Tajik', 'ber-Latn': 'Tamazight', 'ber': 'Tamazight (Tifinagh)', 'ta': 'Tamil', 'tt': 'Tatar', 'te': 'Telugu', 'tet': 'Tetum', 'th': 'Thai', 'bo': 'Tibetan', 'ti': 'Tigrinya', 'tiv': 'Tiv', 'tpi': 'Tok Pisin', 'to': 'Tongan', 'lua': 'Tshiluba', 'ts': 'Tsonga', 'tn': 'Tswana', 'tcy': 'Tulu', 'tum': 'Tumbuka', 'tr': 'Turkish', 'tk': 'Turkmen', 'tyv': 'Tuvan', 'ak': 'Twi', 'udm': 'Udmurt', 'uk': 'Ukrainian', 'ur': 'Urdu', 'ug': 'Uyghur', 'uz': 'Uzbek', 've': 'Venda', 'vec': 'Venetian', 'vi': 'Vietnamese', 'war': 'Waray', 'cy': 'Welsh', 'wo': 'Wolof', 'xh': 'Xhosa', 'sah': 'Yakut', 'yi': 'Yiddish', 'yo': 'Yoruba', 'yua': 'Yucatec Maya', 'zap': 'Zapotec', 'zu': 'Zulu'}


class Google:
    base_url = 'https://translate.google.com'
    translate_rpc_url = base_url + '/_/TranslateWebserverUi/data/batchexecute'

    headers = {
        'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64)',
        'Referer': f'{base_url}/',
    }
    rpc_id = 'MkEWBc'

    def __init__(self):
        self.session = requests.Session()
        self.session.headers.update(self.headers)

    def translate(self, text, src='auto', dst='en'):
        params = {
            'rpcids': self.rpc_id,
            'bl': 'boq_translate-webserver_20201207.13_p0',
            'soc-app': '1',
            'soc-platform': '1',
            'soc-device': '1',
            'rt': 'c',
        }
        data = {
            'f.req': json.dumps(
                [[[
                    self.rpc_id,
                    json.dumps(
                        [[text, src, dst, 1, None, 1], [None]],
                        ensure_ascii=False,
                        separators=(',', ':')
                    ),
                    None,
                    'generic',
                ]]],
                ensure_ascii=False,
                separators=(',', ':'),
            )
        }

        r = self.session.post(
            self.translate_rpc_url + '?' + urllib.parse.urlencode(params),
            data=data,
            timeout=20,
            allow_redirects=False,
        )
        if not r.ok:
            raise ServerException(f'Google Translate returned HTTP {r.status_code}: {r.text[:128]}')

        data = r.content.decode('utf-8')

        json_data = None
        for line in data.split('\n'):
            if self.rpc_id in line:
                json_data = json.loads(line)
                break

        if not json_data:
            raise ServerException('RPC Id payload not found in Google Translate response')

        try:
            payload = json.loads(json_data[0][2])
        except json.JSONDecodeError:
            raise ServerException('Invalid JSON payload in Google Translate response')

        parts = payload[1][0][0][5]
        translated = []
        for index, part in enumerate(parts):
            if not part or not isinstance(part, list) or not isinstance(part[0], str):
                continue

            text = part[0]
            # If part[2] is True, a space (separator) must be added
            if index > 0 and len(part) > 2 and part[2] and text and not text[0].isspace():
                translated.append(' ')

            translated.append(text)

        if not translated:
            raise ServerException('No translated parts found in Google Translate response')

        translated = ''.join(translated)

        src_detected = None
        try:
            src_detected = payload[2]
        except (IndexError, TypeError):
            pass

        origin_pronunciation = None
        try:
            origin_pronunciation = payload[0][0]
        except (IndexError, TypeError):
            pass

        pronunciation = None
        try:
            pronunciation = payload[1][0][0][1]
        except (IndexError, TypeError):
            pass

        mistake = None
        try:
            mistake = payload[0][1][0][0][1]
            # Convert to pango markup
            mistake = mistake.replace('<em>', '<b>').replace('</em>', '</b>')
        except (IndexError, TypeError):
            pass

        return {
            'translated': translated,
            'src_detected': src_detected,
            'origin_pronunciation': origin_pronunciation,
            'pronunciation': pronunciation,
            'mistake': mistake,
        }



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

# Supported servers:
# Aurora Scan [pt_BR] (disabled)
# Cerise Scan [pt_BR] (disabled)
# Dango Scan [pt_BR] (disabled)
# Luratoon Scan [pt_BR] (disabled)
# Nazarick Scan [pt_BR] (disabled)
# RF Dragon Scan [pt_BR] (disabled)
# Wicked Witch Scan [pt_BR] (disabled)

import base64
from io import BytesIO
import re
from zipfile import ZipFile

from bs4 import BeautifulSoup
from PIL import Image
import requests

from komikku.consts import USER_AGENT
from komikku.servers import Server
from komikku.servers.utils import convert_date_string
from komikku.servers.utils import get_soup_element_inner_text
from komikku.utils import get_buffer_mime_type
from komikku.webview import CompleteChallenge

RE_ZIP_IMAGES = re.compile(r'base64,([a-zA-Z0-9+=\/\n]*)')


class Peachscan(Server):
    base_url: str
    search_url: str = None
    latest_updates_url: str = None
    most_populars_url: str = None
    manga_url: str = None
    chapter_url: str = None
    image_url: str = None

    date_format: str = '%d %B %Y'

    def __init__(self):
        if self.session is None and not self.has_cf:
            self.session = requests.Session()
            self.session.headers.update({'User-Agent': USER_AGENT})

        if self.search_url is None:
            self.search_url = self.base_url + '/auto-complete/'
        if self.latest_updates_url is None:
            self.latest_updates_url = self.base_url
        if self.most_populars_url is None:
            self.most_populars_url = self.base_url + '/todas-as-obras/'
        if self.manga_url is None:
            self.manga_url = self.base_url + '/{0}/'
        if self.chapter_url is None:
            self.chapter_url = self.base_url + '/{0}/{1}/'
        if self.image_url is None:
            self.image_url = self.base_url + '{0}#page'

    @CompleteChallenge()
    def get_manga_data(self, initial_data):
        """
        Returns manga data by scraping manga HTML page content

        Initial data should contain at least manga's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Slug is missing in initial data'

        r = self.session_get(self.manga_url.format(initial_data['slug']))
        if r.status_code != 200:
            return None

        data = initial_data.copy()
        data.update(dict(
            authors=[],
            scanlators=[],  # not available
            genres=[],
            status=None,
            synopsis=None,
            chapters=[],
            server_id=self.id,
        ))

        soup = BeautifulSoup(r.text, 'lxml')

        data['name'] = soup.select_one('.desc__titulo__comic').text.strip()
        data['cover'] = self.base_url + soup.select_one('.sumario__img').get('src')

        # Use category as genre
        if element := soup.select_one('.categoria__comic'):
            data['genres'].append(element.text.strip())

        label = None
        for element in soup.select('.sumario__specs .sumario__specs__box'):
            label = element.text.strip()
            value = element.find_next_siblings()[0].text.strip()

            if label.startswith(('Autor',)):
                for author in value.split(','):
                    data['authors'].append(author.strip())

            elif label.startswith(('Gênero',)):
                for genre in value.split():
                    data['genres'].append(genre.strip())

            elif label.startswith(('Status',)):
                if value in ('Em Lançamento',):
                    data['status'] = 'ongoing'
                elif value in ('Finalizado',):
                    data['status'] = 'complete'

        synopsis = []
        for element in soup.select('.sumario__sinopse .sumario__sinopse__texto'):
            synopsis.append(element.text.strip())
        if synopsis:
            data['synopsis'] = '\n\n'.join(synopsis)

        # Chapters
        for element in reversed(soup.select('.link__capitulos')):
            date = None
            if element.parent.name == 'abbr':
                # New chapters are encapsulated in a <abbr> element and date is in `title` attribute
                date = element.parent.get('title')
            elif date_element := element.select_one('.data__lançamento'):
                date = date_element.text

            if date:
                date = date.split()  # Ex. 15 de Maio de 2024 às 16:35
                date = f'{date[0]} {date[2]} {date[4]}'

            slug = element.get('href').split('/')[-2]

            data['chapters'].append(dict(
                slug=slug,
                title=element.select_one('.numero__capitulo').text.strip(),
                num=f'{int(slug)}',
                date=convert_date_string(date, format=self.date_format, languages=[self.lang]),
            ))

        return data

    @CompleteChallenge()
    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        r = self.session_get(self.chapter_url.format(manga_slug, chapter_slug))
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = dict(
            pages=[],
        )

        for script_element in soup.find_all('script'):
            script = script_element.string
            if not script or 'urls' not in script:
                continue

            for line in script.split('\n'):
                for sline in line.split(';'):
                    if 'urls' not in sline:
                        continue

                    urls = sline.split('=')[1].strip()[1:-1].split(',')
                    for index, url in enumerate(urls):
                        data['pages'].append(dict(
                            slug=None,
                            image=url.strip().strip("'"),
                            index=index + 1,
                        ))

                    break

                if data['pages']:
                    break

        return data

    @CompleteChallenge()
    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        r = self.session_get(
            self.image_url.format(page['image']),
            headers={
                'Referer': self.chapter_url.format(manga_slug, chapter_slug),
            }
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type == 'application/zip' or page['image'].endswith('.zip'):
            # Image is cut into several chunks stored in a ZIP archive
            chunks = []
            with BytesIO(r.content) as io_buffer:
                with ZipFile(io_buffer, 'r') as zip:
                    # Read files
                    for name in zip.namelist():
                        _index, type = name.split('.')
                        with zip.open(name=name, mode='r') as fp:
                            data = fp.read()

                        if type == 's':
                            # image data in a SVG
                            data = data.decode('utf-8')
                            if matches := RE_ZIP_IMAGES.search(data):
                                chunks.append(base64.b64decode(matches.group(1)))
                        else:
                            chunks.append(data)

            # Load image chunks and compute its size
            width, height = 0, 0
            img_chunks = []
            for chunk in chunks:
                with BytesIO(chunk) as io_buffer:
                    img_chunk = Image.open(io_buffer)
                img_chunks.append(img_chunk)
                width = max(width, img_chunk.width)
                height += img_chunk.height

            # Put chunks back together
            image = Image.new('RGB', (width, height))
            y = 0
            for img_chunk in img_chunks:
                x = (img_chunk.width - width) / 2
                image.paste(img_chunk, (int(x), y))
                y += img_chunk.height
                img_chunk.close()

            with BytesIO() as io_buffer:
                image.save(io_buffer, 'png')
                buffer = io_buffer.getvalue()

            ext = 'png'

        elif mime_type.startswith('image'):
            buffer = r.content
            ext = mime_type.split('/')[-1]

        else:
            return None

        return dict(
            buffer=buffer,
            mime_type=mime_type,
            name=f'{page["index"]:03d}.{ext}',  # noqa: E231
        )

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return self.manga_url.format(slug)

    @CompleteChallenge()
    def get_latest_updates(self):
        r = self.session_get(self.latest_updates_url)
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        results = []
        for element in soup.select('.all__comics .comic'):
            last_chapter_element = element.select_one('.capitulo__comic')

            results.append({
                'slug': element.a.get('href').split('/')[-2],
                'name': element.h2.text.strip(),
                'cover': self.base_url + element.a.img.get('src'),
                'last_chapter': get_soup_element_inner_text(last_chapter_element, recursive=False).replace('Cap ', '') if last_chapter_element else None,
            })

        return results

    @CompleteChallenge()
    def get_most_populars(self):
        r = self.session_get(self.most_populars_url)
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        results = []
        for element in soup.select('.comics__all__box'):
            img_element = element.select_one('.box-image img')
            a_element = element.select_one('.titulo__comic__allcomics')

            results.append({
                'slug': a_element.get('href').split('/')[-2],
                'name': a_element.text.strip(),
                'cover': self.base_url + img_element.get('src'),
            })

        return results

    @CompleteChallenge()
    def search(self, term):
        r = self.session_get(
            self.search_url,
            params={
                'term': term,
            },
            headers={
                'Referer': f'{self.base_url}/',
            }
        )
        if r.status_code != 200:
            return None

        results = []
        for item in r.json():
            soup = BeautifulSoup(item['html'], 'lxml')
            results.append({
                'slug': soup.a.get('href').split('/')[-2],
                'name': soup.a.span.text.strip(),
                'cover': self.base_url + soup.a.img.get('src'),
            })

        return results



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

import json

from bs4 import BeautifulSoup
import requests

from komikku.consts import USER_AGENT
from komikku.servers import Server
from komikku.servers.utils import convert_date_string
from komikku.utils import get_buffer_mime_type


class Unionmangas(Server):
    id = 'unionmangas'
    name = 'Union Mangás'
    lang = 'pt'
    is_nsfw = True
    long_strip_genres = ['Webtoon', ]
    status = 'disabled'

    base_url = 'https://guimah.com'
    api_search_url = base_url + '/assets/busca.php?nomeManga={0}'
    most_populars_url = base_url + '/lista-mangas/visualizacoes'
    manga_url = base_url + '/perfil/{0}'
    chapter_url = base_url + '/leitor/{0}'
    image_url = 'https://umangas.club/leitor/mangas/{0}/{1}/{2}'

    def __init__(self):
        if self.session is None:
            self.session = requests.Session()
            self.session.headers.update({'user-agent': USER_AGENT})

    def get_manga_data(self, initial_data):
        """
        Returns manga data by scraping manga HTML page content
        """
        r = self.session_get(self.manga_url.format(initial_data['slug']))
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = initial_data.copy()
        data.update(dict(
            authors=[],
            scanlators=[],
            genres=[],
            status=None,
            synopsis=None,
            chapters=[],
            server_id=self.id,
            cover=None,
        ))

        container_element = soup.find('div', class_='perfil-manga')

        data['name'] = container_element.find('h2').text.strip()
        data['cover'] = container_element.find('img', class_='img-thumbnail').get('src')

        for div_element in container_element.find_all('div', class_='col-md-8 col-xs-12'):
            if not div_element.h4:
                continue

            label = div_element.find('label').text.strip()
            div_element.h4.label.extract()
            value = div_element.text.strip()

            if label.startswith('Gênero'):
                data['genres'] = [genre.strip() for genre in value.split(',')]
            elif label.startswith(('Autor', 'Artista')):
                for author in value.split(','):
                    author = author.strip()
                    if author not in data['authors']:
                        data['authors'].append(author)
            elif label.startswith('Status'):
                if value == 'Completo':
                    data['status'] = 'complete'
                elif value == 'Ativo':
                    data['status'] = 'ongoing'

        data['synopsis'] = container_element.find('div', class_='panel-body').text.strip()

        # Chapters
        for div_element in reversed(container_element.find_all('div', class_='row capitulos')):
            a_element = div_element.div.a
            span_element = div_element.div.find_all('span', recursive=False)[1]

            data['chapters'].append(dict(
                title=a_element.text.strip(),
                slug=a_element.get('href').split('/')[-1],
                url='/'.join(a_element.get('href').split('/')[-2:]),
                date=convert_date_string(span_element.text.strip()[1:-1], format='%d/%m/%Y'),
            ))

        return data

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data by scraping chapter HTML page content
        """
        r = self.session_get(self.chapter_url.format(chapter_url))
        if r.status_code != 200:
            return None

        if 'leitor' not in r.url:
            # Chapter page doesn't exist, we have been redirected to manga page
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = dict(
            pages=[],
        )
        for img_element in soup.find_all('img', class_='img-manga'):
            url = img_element.get('src')
            slug = url.split('/')[-1]

            if slug not in ('banner_scan.png', 'banner_forum.png'):
                data['pages'].append(dict(
                    slug=slug,
                    image=None,
                ))
            else:
                data['pages'].append(dict(
                    slug=None,
                    image=url,
                ))

        return data

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        if page['slug']:
            r = self.session_get(self.image_url.format(manga_name, chapter_slug, page['slug']))
        else:
            r = self.session_get(page['image'])
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return dict(
            buffer=r.content,
            mime_type=mime_type,
            name=page['slug'] or page['image'].split('/')[-1],
        )

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return self.manga_url.format(slug)

    def get_most_populars(self):
        r = self.session_get(self.most_populars_url)
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        results = []
        for div_element in soup.find_all('div', class_='lista-mangas-novos'):
            a_element = div_element.find_all('a', recursive=False)[1]

            results.append(dict(
                name=a_element.text.strip(),
                slug=a_element.get('href').split('/')[-1],
            ))

        return results

    def search(self, term=None):
        r = self.session_post(self.api_search_url.format(term))
        if r.status_code != 200:
            return None

        try:
            res = r.json()
        except json.decoder.JSONDecodeError:
            return None

        results = []
        for item in res['items']:
            results.append(dict(
                name=item['titulo'],
                slug=item['url'],
            ))

        return results



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from gettext import gettext as _

from gi.repository import Adw
from gi.repository import GObject
from gi.repository import Gtk

from komikku.models import Category
from komikku.models import CategoryVirtual
from komikku.models import create_db_connection
from komikku.models import Settings


@Gtk.Template.from_resource('/info/febvre/Komikku/ui/categories_editor.ui')
class CategoriesEditorPage(Adw.NavigationPage):
    __gtype_name__ = 'CategoriesEditorPage'

    edited_row = None

    add_entry = Gtk.Template.Child('add_entry')
    add_button = Gtk.Template.Child('add_button')

    stack = Gtk.Template.Child('stack')
    listbox = Gtk.Template.Child('listbox')

    def __init__(self, window):
        Adw.NavigationPage.__init__(self)

        self.window = window

        self.add_entry.connect('activate', self.add_category)
        self.add_button.connect('clicked', self.add_category)

        self.window.navigationview.add(self)

    def add_category(self, _button):
        label = self.add_entry.get_text().strip()
        if not label:
            return

        category = Category.new(label)
        if category:
            self.stack.set_visible_child_name('list')

            self.add_entry.set_text('')
            row = CategoryRow(category)
            row.delete_button.connect('clicked', self.delete_category, row)
            row.save_button.connect('clicked', self.update_category, row)
            row.connect('edit-mode-changed', self.on_category_edit_mode_changed)

            self.listbox.append(row)

            self.window.library.categories_list.populate()

    def delete_category(self, _button, row):
        def confirm_callback():
            deleted_is_current = Settings.get_default().selected_category == row.category.id

            row.category.delete()
            row.category = None
            self.listbox.remove(row)

            if not self.listbox.get_first_child():
                # No more categories
                self.stack.set_visible_child_name('empty')

            # If category is current selected category in Library, reset selected category
            if deleted_is_current:
                Settings.get_default().selected_category = CategoryVirtual.ALL
                self.window.library.flowbox.invalidate_filter()

            self.window.library.categories_list.populate()

        self.window.open_dialog(
            _('Delete?'),
            body=_('Are you sure you want to delete\n"{0}" category?').format(row.category.label),
            confirm_label=_('Delete'),
            confirm_callback=confirm_callback,
            confirm_appearance=Adw.ResponseAppearance.DESTRUCTIVE
        )

    def on_category_edit_mode_changed(self, row, active):
        if not active:
            if self.edited_row == row:
                self.edited_row = None
            return

        if self.edited_row:
            self.edited_row.set_edit_mode(active=False)

        self.edited_row = row

    def populate(self):
        # Clear
        row = self.listbox.get_first_child()
        while row:
            next_row = row.get_next_sibling()
            self.listbox.remove(row)
            row = next_row

        db_conn = create_db_connection()
        records = db_conn.execute('SELECT * FROM categories ORDER BY label ASC').fetchall()

        if records:
            for record in records:
                category = Category.get(record['id'], db_conn=db_conn)

                row = CategoryRow(category)
                row.delete_button.connect('clicked', self.delete_category, row)
                row.save_button.connect('clicked', self.update_category, row)
                row.connect('edit-mode-changed', self.on_category_edit_mode_changed)

                self.listbox.append(row)

            self.stack.set_visible_child_name('list')
        else:
            self.stack.set_visible_child_name('empty')

    def show(self):
        self.populate()

        self.window.navigationview.push(self)

    def update_category(self, _button, row):
        label = row.edit_entry.get_text().strip()
        if not label:
            return

        res = row.category.update(dict(
            label=label,
        ))
        if res:
            row.set_label(label)
            row.set_edit_mode(active=False)

            self.window.library.categories_list.populate()


class CategoryRow(Gtk.ListBoxRow):
    __gsignals__ = {
        'edit-mode-changed': (GObject.SignalFlags.RUN_FIRST, None, (bool, )),
    }

    category = None

    def __init__(self, category):
        Gtk.ListBoxRow.__init__(self, activatable=False)

        self.box = Gtk.Box(
            orientation=Gtk.Orientation.HORIZONTAL, spacing=6, margin_top=8, margin_start=12, margin_bottom=8, margin_end=12
        )

        self.category = category

        self.title_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=12, hexpand=True)

        # Add badge to display number of associated series
        if nb_mangas := len(category.mangas):
            self.badge_label = Gtk.Label(label=str(nb_mangas), valign=Gtk.Align.CENTER)
            self.badge_label.set_css_classes(['badge', 'caption'])
            self.title_box.append(self.badge_label)

        self.title_label = Gtk.Label(label=category.label, hexpand=True, wrap=True, xalign=0)
        self.title_box.append(self.title_label)
        self.box.append(self.title_box)

        self.edit_entry = Gtk.Entry(visible=False, hexpand=True)
        self.edit_entry.set_valign(Gtk.Align.CENTER)
        self.edit_entry.set_halign(Gtk.Align.FILL)
        self.box.append(self.edit_entry)

        self.delete_button = Gtk.Button.new_from_icon_name('user-trash-symbolic')
        self.delete_button.set_tooltip_text(_('Delete'))
        self.delete_button.set_valign(Gtk.Align.CENTER)
        self.delete_button.add_css_class('destructive-action')
        self.box.append(self.delete_button)

        self.edit_button = Gtk.Button.new_from_icon_name('document-edit-symbolic')
        self.edit_button.set_tooltip_text(_('Edit'))
        self.edit_button.set_valign(Gtk.Align.CENTER)
        self.edit_button.connect('clicked', self.set_edit_mode, True)
        self.box.append(self.edit_button)

        self.cancel_button = Gtk.Button.new_from_icon_name('edit-undo-symbolic')
        self.cancel_button.set_tooltip_text(_('Cancel'))
        self.cancel_button.set_valign(Gtk.Align.CENTER)
        self.cancel_button.set_visible(False)
        self.cancel_button.connect('clicked', self.set_edit_mode, False)
        self.box.append(self.cancel_button)

        self.save_button = Gtk.Button.new_from_icon_name('checkmark-symbolic')
        self.save_button.set_tooltip_text(_('Save'))
        self.save_button.set_valign(Gtk.Align.CENTER)
        self.save_button.add_css_class('suggested-action')
        self.save_button.set_visible(False)
        self.box.append(self.save_button)

        self.set_child(self.box)

    def set_edit_mode(self, _button=None, active=False):
        if active:
            self.title_box.set_visible(False)
            self.edit_entry.set_text(self.category.label)
            self.edit_entry.set_visible(True)
            self.delete_button.set_visible(False)
            self.edit_button.set_visible(False)
            self.cancel_button.set_visible(True)
            self.save_button.set_visible(True)
        else:
            self.title_box.set_visible(True)
            self.edit_entry.set_text('')
            self.edit_entry.set_visible(False)
            self.delete_button.set_visible(True)
            self.edit_button.set_visible(True)
            self.cancel_button.set_visible(False)
            self.save_button.set_visible(False)

        self.emit('edit-mode-changed', active)

    def set_label(self, text):
        self.title_label.set_text(text)



# SPDX-FileCopyrightText: 2019-2026 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

import json
import logging
import re

from bs4 import BeautifulSoup
import requests

from komikku.consts import USER_AGENT
from komikku.servers import Server
from komikku.utils import get_buffer_mime_type

logger = logging.getLogger(__name__)


class Lanortrad(Server):
    id = 'lanortrad'
    name = 'LanorTrad'
    lang = 'fr'

    base_url = 'https://lanortrad.com'
    logo_url = base_url + '/images/icons/icon-192x192.png'
    search_url = base_url + '/js/utile/mangaData.js'
    manga_url = base_url + '/manga/{0}'
    chapter_url = base_url + '/manga/{0}/{1}'
    image_url = base_url + '/manga/{0}/{1}'

    def __init__(self):
        if self.session is None:
            self.session = requests.Session()
            self.session.headers = {
                'User-Agent': USER_AGENT,
            }

    def get_manga_data(self, initial_data):
        """
        Returns manga data by scraping manga HTML page content

        Initial data should contain at least manga's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Manga slug is missing in initial data'

        r = self.session_get(self.manga_url.format(initial_data['slug']))
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = initial_data.copy()
        data.update({
            'authors': [],
            'scanlators': [self.name, ],
            'genres': [],
            'status': 'ongoing',
            'synopsis': None,
            'chapters': [],
            'server_id': self.id,
            'cover': None,
        })

        data['name'] = soup.find('h1').text.strip()
        if element := soup.select_one('.card > img'):
            path = element.get('src')
            data['cover'] = f'{self.base_url}/{path}'

        # Details
        for element in soup.select('div.grid div:-soup-contains("Auteur") p, div.grid div:-soup-contains("Artiste") p'):
            author = element.text.strip()
            if author not in data['authors']:
                data['authors'].append(author)

        if element := soup.select_one('div.grid div:-soup-contains("Statut") p'):
            status = element.text.strip()
            if status == 'En cours':
                data['status'] = 'ongoing'
            elif status == 'Terminé':
                data['status'] = 'complete'

        for element in soup.select('div.flex.flex-wrap > span.rounded-full'):
            data['genres'].append(element.text.strip())

        if element := soup.select_one('div.card:-soup-contains("Synopsis") p'):
            data['synopsis'] = element.text.strip()

        # Chapters
        if 'Oneshot' in data['genres']:
            # Oneshot (single chapter)
            data['chapters'].append({
                'slug': 'oneshot',
                'title': 'Chapitre unique',
                'date': None,
            })

        else:
            js_url = None
            for script_element in soup.select('script'):
                if path := script_element.get('src'):
                    if 'js/manga/' in path:
                        path = path.replace('../', '')
                        js_url = f'{self.base_url}/{path}'
                        break

            if js_url is None:
                logger.warning('Failed to get URL of JS file containing manga info')
                return data

            r = self.session_get(
                js_url,
                headers={
                    'Referer': self.manga_url.format(data['slug']),
                }
            )
            if r.status_code != 200:
                return None

            js_code = r.text
            chapters_nums = None
            if matches := re.search(r'\s*maxChapters: (\d*)', js_code):
                chapters_nums = list(range(1, int(matches.group(1)) + 1))

            if chapters_nums is None:
                logger.warning('Failed to retrieve the number of chapters')
                return data

            # Get bonus chapters numbers
            if matches := re.search(r'\s*number: (\d*.\d)', js_code):
                for num in matches.groups():
                    chapters_nums.append(float(num))

            for num in sorted(chapters_nums):
                data['chapters'].append({
                    'slug': f'chapitre {num}',
                    'title': f'Chapitre {num}',
                    'num': num,
                    'date': None,
                })

        return data

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data by scraping chapter HTML page content

        Currently, only pages are expected.
        """
        r = self.session_get(self.chapter_url.format(manga_slug, chapter_slug))
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        nb_digits = None
        nb_pages = None
        path = None
        for script_element in soup.select('script'):
            script = script_element.string
            if not script or 'generateImages' not in script:
                continue

            if matches := re.search(r'\s*padStart\((\d),', script):
                nb_digits = int(matches.group(1))
            if matches := re.search(r'\s*for \(let i = 1; i <= (\d*)\s*; i\+\+\)', script):
                nb_pages = int(matches.group(1))
            if matches := re.search(r'\s*imgElement\.src = `(.*)`', script):
                path = matches.group(1)

            break

        if nb_digits is None:
            logger.warning('Failed to retrieve the number of digits of image filenames')
            return None
        if nb_pages is None:
            logger.warning('Failed to retrieve the number of pages of the chapter')
            return None
        if path is None:
            logger.warning('Failed to retrieve the images path')
            return None

        data = {
            'pages': [],
        }
        for num in range(1, nb_pages + 1):
            data['pages'].append({
                'image': None,
                'url': path.replace('${num}', f'{num:0{nb_digits}}'),  # noqa
            })

        return data

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        r = self.session_get(
            self.image_url.format(manga_slug, page['url']),
            headers={
                'Referer': self.chapter_url.format(manga_slug, chapter_slug),
            }
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return {
            'buffer': r.content,
            'mime_type': mime_type,
            'name': page['url'].split('/')[-1],
        }

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return self.manga_url.format(slug)

    def get_most_populars(self):
        return self.search()

    def search(self, term=None):
        """
        Return all manga by scraping a JS file
        """
        r = self.session_get(
            self.search_url,
            headers={
                'Referer': f'{self.base_url}/',
            }
        )
        if r.status_code != 200:
            return None

        lines = []
        for line in r.text.split('\n'):
            if not line or line.strip().startswith('//'):
                # Empty or comment
                continue

            if line.startswith('window.MANGA_DATA'):
                lines.append('[')
            elif line.strip() == '];':
                lines.append(']')
            elif matches := re.search(r'\s*([a-zA-Z]*): (.*)', line):
                lines.append(f'"{matches.group(1)}": {matches.group(2)}')
            else:
                lines.append(line)

        data = json.loads('\n'.join(lines))

        results = []
        for item in data:
            if term and term.lower() not in item['title'].lower():
                continue

            cover = item['coverImage']
            if not cover.startswith('http'):
                # Oneshots have no cover
                cover = f'{self.base_url}/{item["image"]}'

            results.append({
                'slug': item['id'].lower(),
                'name': item['title'],
                'cover': cover,
            })

        return results



# SPDX-FileCopyrightText: 2019-2026 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

import time

from bs4 import BeautifulSoup
import requests

from komikku.consts import DOWNLOAD_MAX_DELAY
from komikku.consts import USER_AGENT
from komikku.servers import Server
from komikku.servers.utils import convert_date_string
from komikku.utils import get_buffer_mime_type
from komikku.utils import get_response_elapsed


class Iken(Server):
    base_url: str
    api_url: str
    logo_url: str

    manga_url: str = None
    chapter_url: str = None
    api_manga_list_url: str = None
    api_manga_url: str = None
    api_chapters_url: str = None
    api_chapter_url: str = None

    headers: dict = None

    def __init__(self):
        if not self.headers:
            self.headers = {
                'Accept': 'application/json',
                'Origin': self.base_url,
                'User-Agent': USER_AGENT,
            }

        if self.manga_url is None:
            self.manga_url = self.base_url + '/series/{0}'
        if self.chapter_url is None:
            self.chapter_url = self.base_url + '/series/{0}/{1}'
        if self.api_manga_list_url is None:
            self.api_manga_list_url = self.api_url + '/series'
        if self.api_manga_url is None:
            self.api_manga_url = self.api_url + '/series/{0}'
        if self.api_chapters_url is None:
            self.api_chapters_url = self.api_url + '/series/{0}/chapters'
        if self.api_chapter_url is None:
            self.api_chapter_url = self.api_url + '/series/{0}/chapters/{1}'

        if self.session is None:
            self.session = requests.Session()
            self.session.headers.update(self.headers)

    def get_manga_data(self, initial_data):
        """
        Returns manga data via API request

        Initial data should contain at least manga's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Manga slug is missing in initial data'

        r = self.session_get(
            self.api_manga_url.format(initial_data['slug']),
            headers={
                'Referer': f'{self.base_url}/',
            }
        )
        if r.status_code != 200:
            return None

        resp_data = r.json()

        data = initial_data.copy()
        data.update({
            'name': resp_data['title'],
            'authors': [],
            'scanlators': [],
            'genres': [],
            'status': 'ongoing',
            'synopsis': None,
            'chapters': [],
            'server_id': self.id,
            'cover': resp_data['cover'],
        })

        # Details
        for genre in resp_data['genres']:
            data['genres'].append(genre['name'])

        if resp_data['status'] == 'COMPLETED':
            data['status'] = 'complete'
        elif resp_data['status'] == 'ONGOING':
            data['status'] = 'ongoing'
        elif resp_data['status'] == 'DROPPED':
            data['status'] = 'suspended'
        elif resp_data['status'] == 'HIATUS':
            data['status'] = 'hiatus'

        if resp_data.get('author'):
            for author in resp_data['author'].split(','):
                data['authors'].append(author.strip())
        if resp_data.get('artist'):
            for artist in resp_data['artist'].split(','):
                data['authors'].append(artist.strip())

        if synopsis := resp_data['description']:
            data['synopsis'] = BeautifulSoup(synopsis, 'lxml').text.strip()

        # Chapters
        data['chapters'] = self.get_manga_chapters_data(resp_data['slug'])

        return data

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data via API
        """
        r = self.session_get(
            self.api_chapter_url.format(manga_slug, chapter_slug),
            headers={
                'Referer': self.manga_url.format(manga_slug),
            }
        )
        if r.status_code != 200:
            return None

        data = {
            'pages': [],
        }
        for image in r.json()['images']:
            data['pages'].append({
                'slug': None,
                'image': image['url'],
            })

        return data

    def get_manga_chapters_data(self, slug):
        """
        Returns manga chapters list via API
        """
        chapters = []

        def get_page(serie_id, page):
            r = self.session_get(
                self.api_chapters_url.format(slug),
                params={
                    'page': page,
                    'perPage': 30,
                    'sort': 'desc',
                },
                headers={
                    'Referer': f'{self.base_url}/',
                }
            )
            if r.status_code != 200:
                return None, False, None

            data = r.json()
            if not data.get('data'):
                return None, False, None

            more = data['totalPages'] > page

            return data['data'], more, get_response_elapsed(r)

        chapters = []
        delay = None
        more = True
        page = 1
        while more:
            if delay:
                time.sleep(delay)

            chapters_page, more, rtime = get_page(slug, page)
            if chapters_page:
                for chapter in chapters_page:
                    prefix = '🔒 ' if not chapter.get('isFree', True) else ''

                    chapters.append({
                        'slug': chapter['slug'],
                        'title': f'{prefix}Chapter {chapter["number"]}',
                        'num': chapter['number'],
                        'date': convert_date_string(chapter['createdAt'].split('T')[0], '%Y-%m-%d') if 'createdAt' in chapter else None,
                    })
                page += 1
                delay = min(rtime * 2, DOWNLOAD_MAX_DELAY) if rtime else None

            elif chapters_page is None:
                # Failed to retrieve a chapters list page, abort
                break

        return list(reversed(chapters))

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        r = self.session_get(
            page['image'],
            headers={
                'Referer': self.chapter_url.format(manga_slug, chapter_slug),
            }
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return {
            'buffer': r.content,
            'mime_type': mime_type,
            'name': page['image'].split('/')[-1],
        }

    def get_latest_updates(self):
        """
        Returns latest updates
        """
        return self.get_manga_list(orderby='latest')

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return self.manga_url.format(slug)

    def get_manga_list(self, term=None, orderby=None):
        url = self.api_manga_list_url

        params = {
            'page': 1,
            'perPage': 20,
        }
        if term:
            url += '/search'
            params['q'] = term
        else:
            params['sort'] = orderby

        r = self.session_get(
            url,
            params=params,
            headers={
                'Referer': f'{self.base_url}/',
            }
        )
        if r.status_code != 200:
            return None

        results = []
        for item in r.json()['data']:
            results.append({
                'slug': item['slug'],
                'name': item['title'],
                'cover': item['cover'],
            })

        return results

    def get_most_populars(self):
        """
        Returns most popular mangas
        """
        return self.get_manga_list(orderby='popular')

    def search(self, term):
        return self.get_manga_list(term=term)



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

import logging
import re
from urllib.parse import parse_qs
from urllib.parse import urlparse

from bs4 import BeautifulSoup
import requests

from komikku.consts import USER_AGENT
from komikku.servers import Server
from komikku.servers.utils import convert_date_string
from komikku.servers.utils import search_duckduckgo
from komikku.utils import get_buffer_mime_type
from komikku.utils import ServerContent

logger = logging.getLogger('komikku.servers.comicbookplus')


class Comicbookplus(Server):
    id = 'comicbookplus'
    name = 'Comic Book Plus'
    lang = 'en'
    content = ServerContent(
        type=['Collection of free and legal comic books'],
        public_domain=True
    )

    base_url = 'https://comicbookplus.com'
    logo_url = base_url + '/favicon.ico'
    latest_updates_url = base_url + '/?cbplus=latestuploads_s_s_0'
    most_populars_url = base_url + '/?cbplus=mostviewed_s_s_0'
    manga_url = base_url + '/?cid={0}'
    chapter_url = base_url + '/?dlid={0}'

    def __init__(self):
        if self.session is None:
            self.session = requests.Session()
            self.session.headers.update({'user-agent': USER_AGENT})

    @classmethod
    def get_manga_initial_data_from_url(cls, url):
        qs = parse_qs(urlparse(url).query)
        return dict(slug=qs['cid'][0])

    def get_manga_data(self, initial_data):
        """
        Returns comic data by scraping comic HTML page content

        Initial data should contain at least comic's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Comic slug is missing in initial data'

        r = self.session_get(self.manga_url.format(initial_data['slug']))
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = initial_data.copy()
        data.update(dict(
            authors=[],
            scanlators=[],
            genres=[],
            status='complete',
            chapters=[],
            server_id=self.id,
            synopsis=None,
        ))

        cards_elements = soup.select('.introtext')

        data['name'] = soup.select_one('.breadcrumbs > span:last-child').text.strip()
        data['cover'] = cards_elements[0].img.get('src')

        # Details
        if cards_elements[0].table:
            for tr_element in cards_elements[0].table.find_all('tr'):
                td_elements = tr_element.find_all('td')
                label = td_elements[0].text.strip()

                if label.startswith('Categories'):
                    for a_element in td_elements[1].find_all('a'):
                        data['genres'].append(a_element.text.strip())

        if len(cards_elements) == 2:
            data['synopsis'] = cards_elements[1].text.strip()

        # Chapters
        def get_volumes(page=0):
            nonlocal soup

            if page > 0:
                r = self.session_get(self.manga_url.format(initial_data['slug']) + f'&limit={page * 100}')

                soup = BeautifulSoup(r.text, 'lxml')

            chapters_element = soup.find('table', class_='catlistings')
            if chapters_element:
                re_date = r'[a-zA-Z]{3} \d{1,2}, \d{4}'

                for tr_element in chapters_element.select('tr.overrow[itemtype="https://schema.org/Book"]'):
                    date = None
                    for td in tr_element.select('td'):
                        if re.match(re_date, td.text):
                            date = td.text
                        if td.a and td.a.get('itemprop') == 'url':
                            a_element = td.a

                    data['chapters'].append(dict(
                        title=a_element.text.strip(),
                        slug=parse_qs(urlparse(a_element.get('href')).query)['dlid'][0],
                        date=convert_date_string(date, format='%b %d, %Y') if date else None,
                    ))

        get_volumes()

        # Parse next volumes list pages if exist
        if paging_element := soup.select_one('table.catlistings tr.tablefooter'):
            for page in range(1, len(paging_element.find_all('a')) + 1):
                get_volumes(page)

        return data

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url, decode=True):
        """
        Returns comic volume data by scraping volume HTML page content

        Currently, only pages are expected.
        """
        r = self.session_get(self.chapter_url.format(chapter_slug))
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        url = soup.find('img', id='maincomic').get('src')
        split_url = url.split('/')
        base_url = '/'.join(split_url[:-1])
        extension = split_url[-1].split('.')[-1]

        data = dict(
            pages=[],
        )
        for option_element in soup.find('select', id='comicbookpageselect').find_all('option'):
            index = option_element.get('value')
            data['pages'].append(dict(
                slug=None,
                image=f'{base_url}/{index}.{extension}',
            ))

        return data

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns volume page scan (image) content
        """
        r = self.session_get(page['image'])
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return dict(
            buffer=r.content,
            mime_type=mime_type,
            name=page['image'].split('/')[-1],
        )

    def get_manga_url(self, slug, url):
        """
        Returns comic absolute URL
        """
        return self.manga_url.format(slug)

    def get_latest_updates(self):
        """
        Returns latest uploads
        """
        r = self.session_get(self.latest_updates_url)
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        results = []
        cids = []
        for tr_element in soup.find('div', class_='mainbody').find_all('table')[1].find_all('tr')[1:]:
            a_element = tr_element.find_all('td')[1].a
            qs = parse_qs(urlparse(a_element.get('href')).query)

            if qs['cid'][0] in cids:
                continue

            cids.append(qs['cid'][0])
            results.append(dict(
                name=a_element.text.strip(),
                slug=qs['cid'][0],
            ))

        return results

    def get_most_populars(self):
        """
        Returns most viewed comics
        """
        r = self.session_get(self.most_populars_url)
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        results = []
        cids = []
        for tr_element in soup.find('div', class_='mainbody').find('table').find_all('tr')[1:]:
            a_element = tr_element.find_all('td')[1].a
            qs = parse_qs(urlparse(a_element.get('href')).query)

            if qs['cid'][0] in cids:
                continue

            cids.append(qs['cid'][0])
            results.append(dict(
                name=a_element.text.strip(),
                slug=qs['cid'][0],
            ))

        return results

    def search(self, term):
        # Use DuckDuckGo
        results = []
        for ddg_result in search_duckduckgo(urlparse(self.base_url).netloc, term, nb_pages=3):
            qs = parse_qs(urlparse(ddg_result['url']).query)
            if 'cid' not in qs:
                # Not a comic url
                continue

            slug = qs['cid'][0]
            if slug in ('1507',):
                continue
            # Remove ' - Comic Book Plus' at end of name
            name = ddg_result['name'].replace(' - Comic Book Plus', '')

            results.append(dict(
                name=name,
                slug=slug,
            ))

        return results



# SPDX-FileCopyrightText: 2020-2024 GrownNed
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: GrownNed <grownned@gmail.com>

from bs4 import BeautifulSoup
import datetime
import json
import requests

from komikku.servers import Server
from komikku.servers.utils import convert_date_string
from komikku.utils import get_buffer_mime_type
from komikku.utils import is_number


class Readmanga(Server):
    id = 'readmanga'
    name = 'Usagi (Read Manga)'
    lang = 'ru'
    status = 'disabled'

    is_nsfw = True

    base_url = 'https://web.usagi.one'
    logo_url = base_url + '/static/favicon-8d70559a59e754c642d8ccf73bb55d47.ico'
    search_url = base_url + '/search/advancedResults'
    manga_url = base_url + '/{0}'
    chapter_url = manga_url + '/{1}'

    pages_js_start = 'rm_h.readerInit'
    pages_js_offset = 16

    def __init__(self):
        if self.session is None:
            self.session = requests.Session()
            self.session.headers.update({'User-Agent': 'arora'})

    def get_manga_data(self, initial_data):
        """
        Returns manga data by scraping manga HTML page content

        Initial data should contain at least manga's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Manga slug is missing in initial data'

        r = self.session_get(self.manga_url.format(initial_data['slug']))
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = initial_data.copy()
        data.update(dict(
            authors=[],
            scanlators=[],
            genres=[],
            status=None,
            synopsis=None,
            chapters=[],
            server_id=self.id,
        ))

        info_element = soup.select_one('.expandable')

        data['name'] = soup.select_one('meta[itemprop="name"]')['content']
        cover_element = info_element.find('img', attrs={'data-full': True})
        data['cover'] = cover_element.get('data-full')

        # Details
        elements = info_element.find('div', class_='subject-meta').find_all('p', recursive=False)

        if status_element := elements[0].span:
            status = status_element.text.strip()
            if 'продолжается' in status:
                data['status'] = 'ongoing'
            elif 'завершен' in status:
                data['status'] = 'complete'

        for element in elements[1:]:
            label_element = element.span
            if not label_element:
                continue
            label = label_element.text.strip()

            if label.startswith('Автор') or label.startswith('Сценарист') or label.startswith('Художник'):
                value = [author.text.strip() for author in element.find_all('a', class_='person-link')]
                data['authors'].extend(value)
            elif label.startswith('Переводчик'):
                value = [scanlator.text.strip() for scanlator in element.find_all('a', class_='person-link')]
                data['scanlators'].extend(value)
            elif label.startswith('Жанр'):
                value = [genre.text.strip() for genre in element.find_all('a', class_='element-link')]
                data['genres'].extend(value)

        # Synopsis
        data['synopsis'] = soup.select_one('meta[itemprop="description"]')['content']

        # Chapters
        for element in reversed(soup.select('#chapters-list .item-row')):
            a_element = element.find('a', class_='chapter-link')
            slug = a_element.get('href').split('/', 2)[2]
            title = a_element.find(text=True, recursive=False).strip()
            date = element.find('td', class_='text-right').text.strip()

            num = element.get('data-num')
            num = num[:-1] if len(num) > 1 else num  # remove extra `0` at end
            num_volume = element.get('data-vol')

            data['chapters'].append(dict(
                slug=slug,
                title=title,
                num=num if is_number(num) else None,
                num_volume=num_volume if is_number(num_volume) else None,
                date=convert_date_string(date, format='%d.%m.%y'),
            ))

        return data

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data by scraping chapter HTML page content

        Currently, only pages are expected.
        """
        r = self.session_get(self.chapter_url.format(manga_slug, chapter_slug))
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = dict(
            pages=[],
        )

        for script_element in soup.find_all('script'):
            script = script_element.string
            if not script or not script.strip().startswith('var prevLink'):
                continue

            for line in script.split('\n'):
                line = line.strip()
                if not line.startswith(self.pages_js_start):
                    continue

                pages_data = line.split(', ')[0][self.pages_js_offset:]
                pages_data = pages_data.replace("'", '"')
                pages = json.loads(pages_data)

                for split_url in pages:
                    url = split_url[0] + split_url[2]
                    if not url.startswith('http'):
                        # Required by AllHentai
                        url = self.base_url.split('://')[0] + ':' + url

                    data['pages'].append(dict(
                        slug=None,
                        image=url,
                    ))
                break
            break

        return data

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        r = self.session_get(
            page['image'],
            headers={'Referer': f'{self.base_url}/'}
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return dict(
            buffer=r.content,
            mime_type=mime_type,
            name=page['image'].split('/')[-1].split('?')[0],
        )

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return self.manga_url.format(slug)

    def get_most_populars(self):
        """
        Returns best noted manga
        """
        return self.search('', 'POPULARITY')

    def get_latest_updates(self):
        """
        Returns latest updated manga
        """
        return self.search('', 'DATE_UPDATE')

    def search(self, term, orderby='NAME'):
        r = self.session_get(
            self.search_url,
            params=dict(
                q=term,
                offset='',
                years=f'1950,{datetime.date.today().year + 1}',  # noqa: E231
                sortType=orderby,
            )
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type not in ('text/plain', 'text/html'):
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        results = []
        for element in soup.select('.tile'):
            a_element = element.select_one('.desc > h3 > a')
            results.append(dict(
                name=a_element.text.strip(),
                slug=a_element.get('href').split('/')[-1],
                cover=element.select_one('.img img.img-fluid').get('data-original'),
            ))

        return sorted(results, key=lambda m: m['name'])


class Allhentai(Readmanga):
    id = 'allhentai:readmanga'
    name = 'AllHentai'
    is_nsfw_only = True
    # FIXME: requires to be logged in

    base_url = 'https://20.allhen.online'
    logo_url = base_url + '/static/favicon-dcc74ed15e02e1f03481426ba63716e9.ico'
    search_url = base_url + '/search/advanced'
    manga_url = base_url + '/{0}'
    chapter_url = manga_url + '/{1}?mtr=1'


class Mintmanga(Readmanga):
    id = 'mintmanga:readmanga'
    name = 'Mint Manga'
    is_nsfw = True
    status = 'disabled'

    # 16
    base_url = 'https://2.mintmanga.one'
    search_url = base_url + '/search/advancedResults'
    manga_url = base_url + '/{0}'
    chapter_url = manga_url + '/{1}?mtr=1'


class Selfmanga(Readmanga):
    id = 'selfmanga:readmanga'
    name = 'Self Manga'
    is_nsfw = True

    base_url = 'https://selfmanga.live'
    logo_url = base_url + '/static/favicon-29c351d867912538a132f31da35d87aa.ico'
    search_url = base_url + '/search/advancedResults'
    manga_url = base_url + '/{0}'
    chapter_url = manga_url + '/{1}?mtr=1'



# SPDX-FileCopyrightText: 2019-2026 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

import datetime
from gettext import gettext as _
import pytz

from gi.repository import Adw
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk

from komikku.consts import MISSING_IMG_RESOURCE_PATH
from komikku.consts import TIMEZONE
from komikku.models import Chapter
from komikku.models import create_db_connection
from komikku.utils import CoverPicture
from komikku.utils import html_escape

DAYS_LIMIT = 30
THUMB_WIDTH = 45
THUMB_HEIGHT = 62


class HistoryDateBox(Gtk.Box):
    __gtype_name__ = 'HistoryDateBox'

    def __init__(self, window, date, chapters, filter_func):
        super().__init__(orientation=Gtk.Orientation.VERTICAL, spacing=6)

        # Title: Date
        today = datetime.date.today()
        yesterday = today - datetime.timedelta(days=1)
        if date == today:
            label = _('Today')
        elif date == yesterday:
            label = _('Yesterday')
        else:
            label = date.strftime(_('%A, %B %e'))
        date_label = Gtk.Label(label=label, xalign=0)
        date_label.add_css_class('heading')
        self.append(date_label)

        # List of read manga
        self.listbox = Gtk.ListBox()
        self.listbox.add_css_class('boxed-list')
        self.listbox.set_filter_func(filter_func)
        self.append(self.listbox)

        for chapter in chapters:
            self.listbox.append(HistoryDateChapterRow(window, chapter))

    def clear(self):
        for chapter_row in self.listbox:
            chapter_row.clear()


class HistoryDateChapterRow(Adw.ActionRow):
    __gtype_name__ = 'HistoryDateChapterRow'

    def __init__(self, window, chapter):
        super().__init__(activatable=True, selectable=False)

        self.window = window
        self.chapter = chapter

        self.set_title(html_escape(chapter.manga.name))
        self.set_title_lines(1)
        self.set_subtitle(html_escape(chapter.title))
        self.set_subtitle_lines(1)

        # Cover
        if chapter.manga.cover_fs_path is None:
            picture = CoverPicture.new_from_resource(MISSING_IMG_RESOURCE_PATH, THUMB_WIDTH, THUMB_HEIGHT)
        else:
            picture = CoverPicture.new_from_file(chapter.manga.cover_fs_path, THUMB_WIDTH, THUMB_HEIGHT, True)
            if picture is None:
                picture = CoverPicture.new_from_resource(MISSING_IMG_RESOURCE_PATH, THUMB_WIDTH, THUMB_HEIGHT)

        self.cover_frame = Gtk.Frame()
        self.cover_frame.add_css_class('row-rounded-cover-frame')
        self.cover_frame.set_child(picture)
        self.add_prefix(self.cover_frame)

        # Time
        last_read = chapter.last_read.replace(tzinfo=pytz.UTC).astimezone(TIMEZONE)
        label = Gtk.Label(label=last_read.strftime('%H:%M'))
        label.set_css_classes(['subtitle', 'numeric'])
        self.add_suffix(label)

        # Play/Resume button
        self.button = Gtk.Button.new_from_icon_name('media-playback-start-symbolic')
        self.button.set_tooltip_text(_('Resume'))
        self.play_button_clicked_handler_id = self.button.connect('clicked', self.on_play_button_clicked, self)
        self.button.set_valign(Gtk.Align.CENTER)
        self.add_suffix(self.button)

        self.activated_handler_id = self.connect('activated', self.on_activated)

    def clear(self):
        self.button.disconnect(self.play_button_clicked_handler_id)
        self.disconnect(self.activated_handler_id)

    def on_activated(self, row):
        self.window.card.init(row.chapter.manga)

    def on_play_button_clicked(self, _button, row):
        self.window.reader.init(row.chapter.manga, row.chapter)


@Gtk.Template.from_resource('/info/febvre/Komikku/ui/history.ui')
class HistoryPage(Adw.NavigationPage):
    __gtype_name__ = 'HistoryPage'

    search_button = Gtk.Template.Child('search_button')

    stack = Gtk.Template.Child('stack')
    dates_box = Gtk.Template.Child('dates_box')
    searchbar = Gtk.Template.Child('searchbar')
    searchbar_separator = Gtk.Template.Child('searchbar_separator')
    searchentry = Gtk.Template.Child('searchentry')

    def __init__(self, window):
        Adw.NavigationPage.__init__(self)

        self.window = window

        self.connect('hidden', self.on_hidden)

        self.searchbar.bind_property(
            'search-mode-enabled', self.search_button, 'active',
            GObject.BindingFlags.BIDIRECTIONAL | GObject.BindingFlags.SYNC_CREATE
        )
        self.searchbar.bind_property(
            'search-mode-enabled', self.searchbar_separator, 'visible',
            GObject.BindingFlags.BIDIRECTIONAL | GObject.BindingFlags.SYNC_CREATE
        )
        self.searchbar.connect_entry(self.searchentry)
        self.searchbar.set_key_capture_widget(self.window)

        self.searchentry.connect('activate', self.on_searchentry_activated)
        self.searchentry.connect('search-changed', self.search)

        self.window.navigationview.add(self)

    def clear(self):
        date_box = self.dates_box.get_first_child()
        while date_box:
            next_box = date_box.get_next_sibling()
            date_box.clear()
            self.dates_box.remove(date_box)
            date_box = next_box

    def filter(self, row):
        """
        This function gets one row and has to return:
        - True if the row should be displayed
        - False if the row should not be displayed
        """
        term = self.searchentry.get_text().strip().lower()

        ret = (
            term in row.chapter.title.lower() or
            term in row.chapter.manga.name.lower()
        )

        if ret:
            # As soon as a row is visible, made grand parent date_box visible
            GLib.idle_add(row.get_parent().get_parent().set_visible, True)

        return ret

    def on_hidden(self, _page):
        # Leave search mode
        if self.searchbar.get_search_mode():
            self.searchbar.set_search_mode(False)

        # Don't clear on navigation push
        if self.window.previous_page != self.props.tag:
            self.clear()

    def on_searchentry_activated(self, _entry):
        if not self.searchbar.get_search_mode():
            return

        row = self.dates_box.get_first_child().get_last_child().get_row_at_y(0)
        if row:
            self.window.reader.init(row.chapter.manga, row.chapter)

    def populate(self):
        db_conn = create_db_connection()
        start = (datetime.date.today() - datetime.timedelta(days=DAYS_LIMIT)).strftime('%Y-%m-%d')
        query = """
            SELECT DISTINCT manga_id, date(last_read, 'localtime') AS last_read_date, id, max(last_read)
            FROM chapters WHERE date(last_read, 'localtime') >= ?
            GROUP BY manga_id, last_read_date
            ORDER BY last_read DESC
        """
        records = db_conn.execute(query, (start,)).fetchall()

        if records:
            dates = {}
            for record in records:
                chapter = Chapter.get(record['id'], db_conn=db_conn)
                date = record['last_read_date']  # ISO 8601 date string
                if date not in dates:
                    dates[date] = []
                dates[date].append(chapter)

            for iso_date, chapters in dates.items():
                date = datetime.datetime.strptime(iso_date, '%Y-%m-%d').date()
                date_box = HistoryDateBox(self.window, date, chapters, self.filter)
                self.dates_box.append(date_box)

            self.stack.set_visible_child_name('list')
        else:
            self.stack.set_visible_child_name('empty')

    def search(self, _entry):
        for date_box in self.dates_box:
            listbox = date_box.get_last_child()
            listbox.invalidate_filter()
            # Hide date_box, will be shown if a least one row of listbox is not filtered
            date_box.set_visible(False)

    def show(self):
        self.populate()

        self.window.navigationview.push(self)



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from gettext import gettext as _

import requests

from komikku.consts import USER_AGENT
from komikku.servers import Server
from komikku.servers.utils import convert_date_string
from komikku.utils import get_buffer_mime_type
from komikku.utils import ServerContent


class Aralosbd(Server):
    id = 'aralosbd'
    name = 'Aralos BD'
    lang = 'fr'
    content = ServerContent(
        type=[_('Self-publishing platform')]
    )
    long_strip_genres = ['Webtoon']

    base_url = 'https://aralosbd.fr'
    logo_url = base_url + '/icon/favicon-32x32.png'
    search_url = base_url + '/manga/query'
    manga_url = base_url + '/manga/display?id={0}'
    chapter_url = base_url + '/manga/chapter?id={0}'

    api_url = base_url + '/manga/api'
    api_search_url = base_url + '/manga/search'
    latest_updates_url = api_url + '?get=lastchapters&limit=12&order=desc&showReleased=1&showUnreleased=0'
    api_manga_url = api_url + '?get=manga&id={0}'
    api_chapters_url = api_url + '?get=chapters&manga={0}'
    api_chapter_url = api_url + '?get=pages&chapter={0}'

    def __init__(self):
        if self.session is None:
            self.session = requests.Session()
            self.session.headers.update({'User-Agent': USER_AGENT})

    def get_manga_data(self, initial_data):
        """
        Returns manga data via API

        Initial data should contain at least manga's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Manga slug is missing in initial data'

        r = self.session_get(
            self.api_manga_url.format(initial_data['slug']),
            headers={
                'Referer': self.base_url,
            }
        )
        if r.status_code != 200:
            return None

        resp_data = r.json()
        if resp_data['error'] != 0:
            return None

        data = initial_data.copy()
        data.update(dict(
            authors=[author['name'] for author in resp_data['authors']],
            scanlators=[translator['name'] for translator in resp_data['translators']],
            genres=[tag['tag'] for tag in resp_data['tags']],
            status=None,
            synopsis=None,
            chapters=[],
            server_id=self.id,
        ))

        data['name'] = resp_data['main_title']
        data['cover'] = self.base_url + '/' + resp_data['icon']

        if 'Terminé' in data['genres']:
            data['status'] = 'complete'
            data['genres'].remove('Terminé')
        elif 'En cours' in data['genres']:
            data['status'] = 'ongoing'
            data['genres'].remove('En cours')
        elif 'En pause' in data['genres']:
            data['status'] = 'hiatus'
            data['genres'].remove('En pause')

        # Synopsis
        data['synopsis'] = resp_data['description']
        if resp_data['fulldescription']:
            data['synopsis'] += '\n\n' + resp_data['fulldescription']

        # Chapters
        r = self.session_get(
            self.api_chapters_url.format(data['slug']),
            headers={
                'Referer': self.manga_url.format(data['slug']),
            }
        )
        if r.status_code != 200:
            return data

        for chapter in reversed(r.json()):
            data['chapters'].append(dict(
                slug=chapter['chapter_id'],
                title=f"{chapter['chapter_number']} - {chapter['chapter_title']}",
                num=chapter['chapter_number'],
                num_volume=chapter['volume_number'],
                scanlators=[chapter['chapter_translator'], ] if chapter['chapter_translator'] else None,
                date=convert_date_string(chapter['chapter_date'].split()[0], '%Y-%m-%d'),
            ))

        return data

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data via API

        Currently, only pages are expected.
        """
        r = self.session_get(self.api_chapter_url.format(chapter_slug))
        if r.status_code != 200:
            return None

        resp_data = r.json()
        if resp_data['error'] != 0:
            return None

        data = dict(
            pages=[],
        )

        for link in resp_data['links']:
            data['pages'].append(dict(
                slug=None,
                image=link,
            ))

        return data

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        r = self.session_get(
            f"{self.base_url}/{page['image']}",
            headers={
                'Referer': self.chapter_url.format(chapter_slug),
            }
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return dict(
            buffer=r.content,
            mime_type=mime_type,
            name=f"{page['image'].split('?')[0].split('/')[-1]}.{mime_type.split('/')[-1]}",
        )

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return self.manga_url.format(slug)

    def get_latest_updates(self):
        """
        Returns list of latest mangas
        """
        r = self.session_get(
            self.latest_updates_url,
            headers={
                'Referer': self.base_url,
            }
        )
        if r.status_code != 200:
            return None

        resp_data = r.json()
        if resp_data['error'] != 0:
            return None

        results = {}
        for chapter in resp_data['chapters']:
            if chapter['manga_id'] not in results:
                results[chapter['manga_id']] = dict(
                    slug=chapter['manga_id'],
                    name=chapter['manga_title'],
                    cover=self.base_url + '/' + chapter['icon'],
                    last_chapter=chapter['chapter_number'],
                )
            elif chapter['chapter_number'] > results[chapter['manga_id']]['last_chapter']:
                results[chapter['manga_id']]['last_chapter'] = chapter['chapter_number']

        return results.values()

    def get_most_populars(self):
        """
        Returns list of most viewed mangas
        """
        return self.search('', populars=True)

    def search(self, term, populars=False):
        if populars:
            filters = 'sort:allviews;limit:18;-id:3;order:desc'
        else:
            filters = f'title~{term};limit:18;order:asc'  # noqa: E231, E702

        def get_page(num=0):
            r = self.session_get(
                self.api_search_url,
                params=dict(
                    s=f'{filters};page:{num}',  # noqa: E231, E702
                ),
                headers={
                    'Referer': self.search_url,
                }
            )
            if r.status_code != 200:
                return None

            data = r.json()
            if data['error'] != 0:
                return None

            return data

        results = []
        more = True
        page_num = -1
        while more:
            page_num += 1
            data = get_page(page_num)
            if data is None:
                break

            for manga in data['mangas']:
                results.append(dict(
                    slug=manga['id'],
                    name=manga['title'],
                    cover=self.base_url + '/' + manga['icon'],
                    nb_chapters=manga['chapter_count'],
                ))

            more = page_num < data['page_count'] - 1

        return results



# SPDX-FileCopyrightText: 2019-2026 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from gettext import gettext as _

from gi.repository import Adw
from gi.repository import Gdk
from gi.repository import Gio
from gi.repository import GLib
from gi.repository import Gtk

from komikku.card.categories_list import CategoriesList
from komikku.card.chapters_list import ChaptersList
from komikku.card.info_box import InfoBox
from komikku.card.tracking import TrackingDialog
from komikku.models import Settings
from komikku.utils import html_escape


@Gtk.Template.from_resource('/info/febvre/Komikku/ui/card.ui')
class CardPage(Adw.NavigationPage):
    __gtype_name__ = 'CardPage'

    toolbar_view = Gtk.Template.Child('toolbar_view')
    left_button = Gtk.Template.Child('left_button')
    filters_button = Gtk.Template.Child('filters_button')
    title_stack = Gtk.Template.Child('title_stack')
    title = Gtk.Template.Child('title')
    viewswitcher = Gtk.Template.Child('viewswitcher')
    menu_button = Gtk.Template.Child('menu_button')

    activity_progressbar = Gtk.Template.Child('activity_progressbar')
    stack = Gtk.Template.Child('stack')
    categories_stack = Gtk.Template.Child('categories_stack')
    categories_scrolledwindow = Gtk.Template.Child('categories_scrolledwindow')
    categories_listbox = Gtk.Template.Child('categories_listbox')
    chapters_scrolledwindow = Gtk.Template.Child('chapters_scrolledwindow')
    chapters_listview = Gtk.Template.Child('chapters_listview')
    chapters_selection_mode_actionbar = Gtk.Template.Child('chapters_selection_mode_actionbar')
    chapters_selection_mode_download_button = Gtk.Template.Child('chapters_selection_mode_download_button')
    chapters_selection_mode_clear_button = Gtk.Template.Child('chapters_selection_mode_clear_button')
    chapters_selection_mode_clear_reset_button = Gtk.Template.Child('chapters_selection_mode_clear_reset_button')
    chapters_selection_mode_menubutton = Gtk.Template.Child('chapters_selection_mode_menubutton')
    info_scrolledwindow = Gtk.Template.Child('info_scrolledwindow')
    backdrop_picture = Gtk.Template.Child('backdrop_picture')
    title_box = Gtk.Template.Child('title_box')
    cover_box = Gtk.Template.Child('cover_box')
    name_label = Gtk.Template.Child('name_label')
    authors_label = Gtk.Template.Child('authors_label')
    status_server_label = Gtk.Template.Child('status_server_label')
    buttons_box = Gtk.Template.Child('buttons_box')
    add_button = Gtk.Template.Child('add_button')
    resume_button = Gtk.Template.Child('resume_button')
    genres_wrapbox = Gtk.Template.Child('genres_wrapbox')
    categories_wrapbox = Gtk.Template.Child('categories_wrapbox')
    scanlators_label = Gtk.Template.Child('scanlators_label')
    chapters_label = Gtk.Template.Child('chapters_label')
    last_update_label = Gtk.Template.Child('last_update_label')
    synopsis_box = Gtk.Template.Child('synopsis_box')
    synopsis_togglebutton = Gtk.Template.Child('synopsis_togglebutton')
    size_on_disk_label = Gtk.Template.Child('size_on_disk_label')
    viewswitcherbar = Gtk.Template.Child('viewswitcherbar')

    manga = None
    pool_to_update = False
    pool_to_update_offset = 0
    selection_mode = False

    def __init__(self, window):
        Adw.NavigationPage.__init__(self)

        self.window = window
        self.builder = window.builder
        self.builder.add_from_resource('/info/febvre/Komikku/ui/menu/card.xml')
        self.builder.add_from_resource('/info/febvre/Komikku/ui/menu/card_selection_mode.xml')

        self.css_provider = Gtk.CssProvider.new()
        Gtk.StyleContext.add_provider_for_display(Gdk.Display.get_default(), self.css_provider, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)

        self.connect('shown', self.on_shown)
        self.stack.connect('notify::visible-child-name', self.on_page_changed)
        self.window.controller_key.connect('key-pressed', self.on_key_pressed)

        # Header bar
        self.left_button.connect('clicked', self.leave_selection_mode)
        self.filters_button.connect('clicked', self.on_filters_button_clicked)
        self.menu_button.set_menu_model(self.builder.get_object('menu-card'))
        # Focus is lost after showing popover submenu (bug?)
        self.menu_button.get_popover().connect('closed', lambda _popover: self.menu_button.grab_focus())

        # Pool-to-Update
        self.pool_to_update_revealer = self.window.pool_to_update_revealer
        self.pool_to_update_spinner = self.window.pool_to_update_spinner
        # Drag gesture
        self.gesture_drag = Gtk.GestureDrag.new()
        self.gesture_drag.set_touch_only(True)
        self.gesture_drag.set_propagation_phase(Gtk.PropagationPhase.CAPTURE)
        self.gesture_drag.connect('drag-end', self.on_gesture_drag_end)
        self.gesture_drag.connect('drag-update', self.on_gesture_drag_update)
        self.stack.add_controller(self.gesture_drag)

        self.tracking_dialog = TrackingDialog(self.window)

        self.info_box = InfoBox(self)
        self.categories_list = CategoriesList(self)
        self.chapters_list = ChaptersList(self)

        self.filters_dialog = Adw.PreferencesDialog.new()
        self.filters_dialog.set_title(_('Filters'))
        self.filters_dialog.props.presentation_mode = Adw.DialogPresentationMode.BOTTOM_SHEET

        self.window.updater.connect('manga-updated', self.on_manga_updated)
        self.window.trackers.connect('manga-tracker-synced', self.on_manga_tracker_synced)

        self.window.breakpoint.add_setter(self.viewswitcherbar, 'reveal', True)
        self.window.breakpoint.add_setter(self.title_stack, 'visible-child', self.title)

        self.window.navigationview.add(self)

    def add_actions(self):
        self.resume_action = Gio.SimpleAction.new('card.resume', None)
        self.resume_action.connect('activate', self.on_resume_button_clicked)
        self.window.application.add_action(self.resume_action)

        self.update_action = Gio.SimpleAction.new('card.update', None)
        self.update_action.connect('activate', self.on_update_request)
        self.window.application.add_action(self.update_action)

        self.clear_action = Gio.SimpleAction.new('card.clear', None)
        self.clear_action.connect('activate', self.on_clear_menu_clicked)
        self.window.application.add_action(self.clear_action)

        self.delete_action = Gio.SimpleAction.new('card.delete', None)
        self.delete_action.connect('activate', self.on_delete_menu_clicked)
        self.window.application.add_action(self.delete_action)

        variant = GLib.Variant.new_string('desc')
        self.sort_order_action = Gio.SimpleAction.new_stateful('card.sort-order', variant.get_type(), variant)
        self.sort_order_action.connect('activate', self.chapters_list.on_sort_order_changed)
        self.window.application.add_action(self.sort_order_action)

        self.manage_tracking_action = Gio.SimpleAction.new('card.manage-tracking', None)
        self.manage_tracking_action.connect('activate', self.on_manage_tracking_menu_clicked)
        self.window.application.add_action(self.manage_tracking_action)

        self.open_in_browser_action = Gio.SimpleAction.new('card.open-in-browser', None)
        self.open_in_browser_action.connect('activate', self.on_open_in_browser_menu_clicked)
        self.window.application.add_action(self.open_in_browser_action)

        self.chapters_list.add_actions()

    def enter_selection_mode(self, init=False):
        if self.selection_mode:
            return

        self.selection_mode = True
        self.chapters_list.enter_selection_mode(init)

        self.props.can_pop = False

        self.left_button.set_label(_('Cancel'))
        self.left_button.set_tooltip_text(_('Cancel'))
        self.left_button.set_visible(True)

        self.viewswitcher.set_visible(False)
        self.viewswitcherbar.set_visible(False)
        self.title_stack.set_visible_child(self.title)

        self.menu_button.set_visible(False)

    def init(self, manga):
        # Default page is `Info`
        self.stack.set_visible_child_name('info')

        # Hide Categories if manga is not in Library
        self.stack.get_page(self.stack.get_child_by_name('categories')).set_visible(manga.in_library)

        self.manga = manga
        # Unref chapters to force a reload
        self.manga._chapters = None

        self.set_unread_chapters_badge()

        self.show()

    def leave_selection_mode(self, *args):
        self.selection_mode = False
        self.chapters_list.leave_selection_mode()

        self.props.can_pop = True

        self.left_button.set_visible(False)

        self.viewswitcher.set_visible(True)
        self.viewswitcherbar.set_visible(True)
        self.title.set_subtitle('')
        if not self.viewswitcherbar.get_reveal():
            self.title_stack.set_visible_child(self.viewswitcher)

        self.menu_button.set_visible(True)

    def on_add_button_clicked(self, _button):
        # Show categories
        self.stack.get_page(self.stack.get_child_by_name('categories')).set_visible(True)
        # Hide Add to Library button
        self.info_box.add_button.set_visible(False)
        self.info_box.resume_button.add_css_class('suggested-action')
        # Update manga
        self.manga.add_in_library()
        self.window.library.on_manga_added(self.manga)

    def on_clear_menu_clicked(self, _action, _gparam):
        def confirm_callback():
            self.manga.clear()
            self.refresh(info=True, chapters=self.manga.chapters)
            self.window.library.refresh_on_manga_state_changed(self.manga)

        self.window.open_dialog(
            _('Clear?'),
            body=_('Are you sure you want to clear all chapters?'),
            confirm_label=_('Clear'),
            confirm_callback=confirm_callback,
            confirm_appearance=Adw.ResponseAppearance.DESTRUCTIVE
        )

    def on_delete_menu_clicked(self, _action, _gparam):
        self.window.library.delete_mangas([self.manga, ])

    def on_filters_button_clicked(self, _button):
        filters = self.manga.filters or {}

        def on_active(row, _param, scanlator):
            if 'scanlators' not in filters:
                filters['scanlators'] = []

            if row.get_active():
                filters['scanlators'].remove(scanlator)
            else:
                filters['scanlators'].append(scanlator)

            if filters['scanlators']:
                self.filters_button.add_css_class('accent')
            else:
                self.filters_button.remove_css_class('accent')

            self.manga.update({
                'filters': filters,
            })

            self.chapters_list.list_model.invalidate_filter()

        # Remove a previous used page if exists
        if page := self.filters_dialog.get_visible_page():
            self.filters_dialog.remove(page)

        page = Adw.PreferencesPage(title=_('Filters'))
        self.filters_dialog.add(page)

        group = Adw.PreferencesGroup()
        group.set_title('Chapters Scanlation Groups')

        for scanlator in self.manga.chapters_scanlators:
            title = html_escape(scanlator['name'])
            if title == 'Unknown':
                title = _('Unknown')

            row = Adw.SwitchRow(title=title)
            row.set_use_markup(True)
            row.set_active(not filters or 'scanlators' not in filters or scanlator['name'] not in filters['scanlators'])
            row.connect('notify::active', on_active, scanlator['name'])

            label = Gtk.Label(label=str(scanlator['count']), valign=Gtk.Align.CENTER)
            label.set_css_classes(['badge', 'caption'])
            row.add_prefix(label)

            group.add(row)

        page.add(group)

        self.filters_dialog.present(self.window)

    def on_gesture_drag_end(self, _controller, _offset_x, _offset_y):
        self.stack.set_opacity(1)
        self.stack.remove_css_class('grayscale')

        if not self.pool_to_update:
            return

        self.pool_to_update = False
        self.pool_to_update_revealer.props.transition_type = Gtk.RevealerTransitionType.SLIDE_DOWN
        self.pool_to_update_revealer.set_reveal_child(False)
        self.pool_to_update_spinner.props.margin_top = 0

        if self.pool_to_update_offset > 2 * 150:
            self.on_update_request()

        self.gesture_drag.set_state(Gtk.EventSequenceState.CLAIMED)

    def on_gesture_drag_update(self, _controller, _offset_x, offset_y):
        _active, start_x, start_y = self.gesture_drag.get_start_point()

        page = self.stack.get_visible_child_name()
        if page == 'info':
            scroll_value = self.info_scrolledwindow.get_vadjustment().props.value
        elif page == 'chapters':
            scroll_value = self.chapters_scrolledwindow.get_vadjustment().props.value
        else:
            scroll_value = self.categories_scrolledwindow.get_vadjustment().props.value

        if scroll_value != 0 or offset_y < 0 or self.selection_mode or start_x < 32 or start_y > self.get_height() / 3:
            return

        self.stack.set_opacity(0.5)
        self.stack.add_css_class('grayscale')

        self.pool_to_update_offset = offset_y

        if not self.pool_to_update:
            self.pool_to_update = True
            # Adjust revealer position
            self.pool_to_update_revealer.set_margin_top(self.toolbar_view.get_top_bar_height())
            self.pool_to_update_revealer.props.transition_type = Gtk.RevealerTransitionType.NONE
            self.pool_to_update_revealer.set_reveal_child(True)
        else:
            self.pool_to_update_spinner.props.margin_top = max(0, min(150, offset_y / 2))

    def on_key_pressed(self, _controller, keyval, _keycode, state):
        if self.window.page != self.props.tag:
            return Gdk.EVENT_PROPAGATE
        if self.stack.get_visible_child_name() != 'chapters':
            return Gdk.EVENT_PROPAGATE

        modifiers = state & Gtk.accelerator_get_default_mod_mask()
        if self.selection_mode:
            if keyval == Gdk.KEY_Escape or (modifiers == Gdk.ModifierType.ALT_MASK and keyval in (Gdk.KEY_Left, Gdk.KEY_KP_Left)):
                self.leave_selection_mode()
                # Stop event to prevent back navigation
                return Gdk.EVENT_STOP
        else:
            # Allow to enter in selection mode with <SHIFT>+Arrow key
            if modifiers != Gdk.ModifierType.SHIFT_MASK or keyval not in (Gdk.KEY_Up, Gdk.KEY_KP_Up, Gdk.KEY_Down, Gdk.KEY_KP_Down):
                return Gdk.EVENT_PROPAGATE

            self.enter_selection_mode()

        return Gdk.EVENT_PROPAGATE

    def on_manage_tracking_menu_clicked(self, _action, _menu):
        self.tracking_dialog.show()

    def on_manga_tracker_synced(self, _trackers, manga):
        if (self.window.page == self.props.tag or self.window.previous_page == self.props.tag) and self.manga.id == manga.id:
            self.manga = manga

    def on_manga_updated(self, _updater, manga, chapters_changes, synced):
        if (self.window.page == self.props.tag or self.window.previous_page == self.props.tag) and self.manga.id == manga.id:
            self.manga = manga

            self.info_box.populate()
            self.toggle_filters_button()

            if sum(chapters_changes.values()) > 0:
                self.chapters_list.populate()
                self.refresh(unread_chapters=True)

            if synced:
                self.window.add_notification(_('Read progress synchronization with server completed successfully'))

    def on_open_in_browser_menu_clicked(self, _action, _gparam):
        if uri := self.manga.server.get_manga_url(self.manga.slug, self.manga.url):
            Gtk.UriLauncher.new(uri=uri).launch()
        else:
            self.window.add_notification(_('Failed to get manga URL'))

    def on_page_changed(self, _page, _gparam):
        self.toggle_filters_button()

    def on_resume_button_clicked(self, *args):
        last_read_chapter = self.manga.last_read_chapter
        if last_read_chapter and last_read_chapter.read:
            # If last read chapter has been read in full, use next chapter
            last_read_chapter = self.manga.get_next_chapter(last_read_chapter)

        if last_read_chapter is None:
            # Use first chapter
            if self.chapters_list.sort_order.endswith('desc'):
                last_read_chapter = self.chapters_list.list_model.get_item(self.chapters_list.list_model.get_n_items() - 1).chapter
            else:
                last_read_chapter = self.chapters_list.list_model.get_item(0).chapter

        self.window.reader.init(self.manga, last_read_chapter)

    def on_shown(self, _page):
        def do_populate():
            if self.manga.server.status == 'disabled':
                self.window.add_notification(
                    _('NOTICE\n{0} is no longer supported.\nPlease switch to another server.').format(self.manga.server.name)
                )

            # Show update indicator (in case an update is in progress)
            self.show_activity_indicator()

            if self.window.last_navigation_action != 'push':
                return

            # Wait page is shown (transition is ended) to populate
            # Operation is resource intensive and could disrupt page transition
            self.populate()

        if not Gtk.Settings.get_default().get_property('gtk-enable-animations'):
            # When animations are disabled, popped/pushed events are sent after `shown` event (bug?)
            # Use idle_add to be sure that last `popped` or `pushed` event has been received
            GLib.idle_add(do_populate)
        else:
            do_populate()

    def on_update_request(self, _action=None, _param=None):
        self.window.updater.add(self.manga)
        if self.window.updater.start():
            # Start update indicator
            self.show_activity_indicator()

    def populate(self):
        self.chapters_list.set_sort_order(invalidate=False)

        self.chapters_list.populate()
        self.categories_list.populate()

    def refresh(self, unread_chapters=False, info=False, chapters=None):
        if unread_chapters:
            self.set_unread_chapters_badge()

        if info:
            self.info_box.refresh()

        if chapters is not None:
            self.chapters_list.refresh(chapters)

    def remove_backdrop(self):
        self.remove_css_class('backdrop')
        self.remove_css_class('backdrop-linear-gradient')
        self.backdrop_picture.set_paintable(None)
        self.backdrop_picture.set_opacity(1)

    def set_actions_enabled(self, enabled):
        self.delete_action.set_enabled(enabled)
        self.update_action.set_enabled(enabled)
        self.sort_order_action.set_enabled(enabled)

    def set_backdrop(self):
        self.remove_backdrop()

        method = Settings.get_default().card_backdrop_method
        if not self.manga or Adw.StyleManager.get_default().get_high_contrast() or not method:
            return

        if method == 'blurred-cover':
            if path := self.manga.backdrop_image_fs_path:
                self.backdrop_picture.set_filename(path)

                if info := self.manga.backdrop_info:
                    if Adw.StyleManager.get_default().get_dark():
                        opacity = 1 - info['luminance'][0]
                    else:
                        opacity = info['luminance'][1]
                    self.backdrop_picture.set_opacity(opacity)

        elif method == 'linear-gradient':
            if css := self.manga.backdrop_colors_css:
                self.css_provider.load_from_string(css)
                self.add_css_class('backdrop-linear-gradient')

        self.add_css_class('backdrop')

    def set_unread_chapters_badge(self):
        # Show unread chapters (with Adw.ViewStackPage badge) if any
        if unread_chapters := self.manga.nb_unread_chapters:
            self.stack.get_page(self.stack.get_child_by_name('chapters')).set_badge_number(unread_chapters)
        else:
            self.stack.get_page(self.stack.get_child_by_name('chapters')).set_badge_number(0)

    def show(self):
        self.props.title = self.manga.name  # Adw.NavigationPage title

        self.toggle_filters_button()
        self.title.set_title(self.manga.name)
        self.info_box.populate()

        # Adjust some buttons of action bar (visible in selection mode)
        self.chapters_selection_mode_download_button.set_sensitive(not self.manga.is_local)
        self.chapters_selection_mode_clear_button.set_sensitive(not self.manga.is_local)
        self.chapters_selection_mode_clear_reset_button.set_tooltip_text(_('Clear and Reset') if not self.manga.is_local else _('Reset'))

        self.open_in_browser_action.set_enabled(not self.manga.is_local)

        # Reset scrolling in all pages
        for page in self.stack.get_pages():
            scrolledwindow = page.get_child() if page.props.name != 'chapters' else page.get_child().get_first_child()
            scrolledwindow.get_vadjustment().configure(0, 0, 0, 0, 0, 0)

        self.window.navigationview.push(self)

    def show_activity_indicator(self):
        def pulse(manga_id):
            if self.window.page != self.props.tag:
                # Page left, stop indicator
                self.activity_progressbar.props.fraction = 0
                return GLib.SOURCE_REMOVE

            if self.window.updater.current_id == manga_id:
                self.activity_progressbar.pulse()
                return GLib.SOURCE_CONTINUE

            # Update is ended, stop indicator
            self.activity_progressbar.props.fraction = 0
            return GLib.SOURCE_REMOVE

        GLib.timeout_add(250, pulse, self.manga.id)

    def toggle_filters_button(self):
        name = self.stack.get_visible_child_name()

        if name == 'chapters':
            # Show button if chapters have at least 2 different scanlators
            self.filters_button.set_visible(self.manga.chapters_scanlators and len(self.manga.chapters_scanlators) > 1)
            if self.manga.filters and self.manga.filters.get('scanlators'):
                self.filters_button.add_css_class('accent')
            else:
                self.filters_button.remove_css_class('accent')
        else:
            # Button is visible in `Chapters` page only
            self.filters_button.set_visible(False)

    def toggle_resume(self, state):
        self.resume_action.set_enabled(state)
        self.resume_button.set_sensitive(state)



# SPDX-FileCopyrightText: 2019-2026 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from gettext import gettext as _

from gi.repository import Adw
from gi.repository import Gdk
from gi.repository import Gio
from gi.repository import GObject
from gi.repository import Gtk

from komikku.consts import BORDERS_CROP_THRESHOLDS
from komikku.models import Settings


class KeyLabelPair(GObject.Object):
    key = GObject.Property(type=str, flags=GObject.ParamFlags.READWRITE, default='')
    label = GObject.Property(type=str, nick='Label', blurb='Label', flags=GObject.ParamFlags.READWRITE, default='')


@Gtk.Template.from_resource('/info/febvre/Komikku/ui/reader_settings.ui')
class ReaderSettingsDialog(Adw.PreferencesDialog):
    __gtype_name__ = 'ReaderSettingsDialog'

    page_filters_group = Gtk.Template.Child('page_filters_group')
    scaling_row = Gtk.Template.Child('scaling_row')
    scaling_filter_row = Gtk.Template.Child('scaling_filter_row')
    background_color_row = Gtk.Template.Child('background_color_row')
    landscape_zoom_switch = Gtk.Template.Child('landscape_zoom_switch')
    borders_crop_switch = Gtk.Template.Child('borders_crop_switch')
    borders_crop_threshold_row = Gtk.Template.Child('borders_crop_threshold_row')
    page_numbering_switch = Gtk.Template.Child('page_numbering_switch')
    ocr_lang_row = Gtk.Template.Child('ocr_lang_row')

    def __init__(self, reader):
        Adw.PreferencesDialog.__init__(self)

        self.reader = reader
        self.settings = Settings.get_default()

        self.css_provider = Gtk.CssProvider.new()
        Gtk.StyleContext.add_provider_for_display(
            Gdk.Display.get_default(), self.css_provider, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
        )

        #
        # Custom settings
        #
        list_store_expression = Gtk.PropertyExpression.new(KeyLabelPair, None, 'label')

        # Scaling
        self.scalings = {
            'screen': _('Adapt to Screen'),
            'width': _('Adapt to Width'),
            'height': _('Adapt to Height'),
            'original': _('Original Size'),
        }

        model = Gio.ListStore(item_type=KeyLabelPair)
        items = [KeyLabelPair(key=key, label=label) for key, label in self.scalings.items()]
        model.splice(0, 0, items)
        self.scaling_row.set_expression(list_store_expression)
        self.scaling_row.set_model(model)

        # Scaling filter
        self.scaling_filters = {
            'linear': _('Linear'),
            'trilinear': _('Trilinear'),
        }

        model = Gio.ListStore(item_type=KeyLabelPair)
        items = [KeyLabelPair(key=key, label=label) for key, label in self.scaling_filters.items()]
        model.splice(0, 0, items)
        self.scaling_filter_row.set_expression(list_store_expression)
        self.scaling_filter_row.set_model(model)

        # Background color
        self.background_colors = {
            'white': _('White'),
            'gray': _('Gray'),
            'black': _('Black'),
            'system-style': _('System Style'),
        }

        model = Gio.ListStore(item_type=KeyLabelPair)
        items = [KeyLabelPair(key=key, label=label) for key, label in self.background_colors.items()]
        model.splice(0, 0, items)
        self.background_color_row.set_expression(list_store_expression)
        self.background_color_row.set_model(model)

        # Borders crop threshold
        model = Gio.ListStore(item_type=KeyLabelPair)
        items = [KeyLabelPair(key=str(value), label=str(value)) for value in BORDERS_CROP_THRESHOLDS]
        model.splice(0, 0, items)
        self.borders_crop_threshold_row.set_expression(list_store_expression)
        self.borders_crop_threshold_row.set_model(model)

        # OCR language
        if self.reader.ocr_translator.enabled:
            model = Gio.ListStore(item_type=KeyLabelPair)
            items = [KeyLabelPair(key=value, label=value) for value in self.reader.ocr_translator.ocr_languages]
            model.splice(0, 0, items)
            self.ocr_lang_row.set_expression(list_store_expression)
            self.ocr_lang_row.set_model(model)

        self.init_general()

    def init_custom(self):
        # Scaling
        self.scaling_row.set_selected(list(self.scalings.keys()).index(self.reader.scaling))
        self.scaling_row.connect('notify::selected', self.on_scaling_changed)

        # Scaling filter
        self.scaling_filter_row.set_selected(list(self.scaling_filters.keys()).index(self.reader.scaling_filter))
        self.scaling_filter_row.connect('notify::selected', self.on_scaling_filter_changed)

        # Background color
        self.background_color_row.set_selected(list(self.background_colors.keys()).index(self.reader.background_color))
        self.background_color_row.connect('notify::selected', self.on_background_color_changed)

        # Landscape zoom
        self.landscape_zoom_switch.set_active(self.reader.landscape_zoom)
        self.landscape_zoom_switch.connect('notify::active', self.on_landscape_zoom_changed)

        # Borders crop
        self.borders_crop_switch.set_active(self.reader.borders_crop)
        self.borders_crop_switch.connect('notify::active', self.on_borders_crop_changed)

        # Borders crop threshold
        self.borders_crop_threshold_row.set_selected(BORDERS_CROP_THRESHOLDS.index(self.reader.borders_crop_threshold))
        self.borders_crop_threshold_row.connect('notify::selected', self.on_borders_crop_threshold_changed)

        # Page numbering
        self.page_numbering_switch.set_active(not self.reader.page_numbering)
        self.page_numbering_switch.connect('notify::active', self.on_page_numbering_changed)

        # OCR language
        if self.reader.ocr_translator.enabled:
            self.ocr_lang_row.set_selected(self.reader.ocr_translator.ocr_languages.index(self.reader.ocr_translator.ocr_lang))
            self.ocr_lang_row.connect('notify::selected', self.on_ocr_lang_changed)
        else:
            self.ocr_lang_row.set_sensitive(False)

    def init_general(self):
        # Filters
        filters = {
            'brightness': {
                'title': _('Brightness'),
                'subtitle': _('Make pages brighter or darker'),
                'min': 0,
                'max': 200,
                'step': 1,
                'default': 100,
            },
            'contrast': {
                'title': _('Contrast'),
                'subtitle': _('Increase or decrease the contrast of pages'),
                'min': 0,
                'max': 200,
                'step': 1,
                'default': 100,
            },
            'grayscale': {
                'title': _('Grayscale'),
                'subtitle': _('Convert pages to grayscale'),
                'min': 0,
                'max': 100,
                'step': 1,
                'default': 0,
            },
            'sepia': {
                'title': _('Sepia'),
                'subtitle': _('Give a more yellow/brown appearance to pages'),
                'min': 0,
                'max': 100,
                'step': 1,
                'default': 0,
            },
            'saturate': {
                'title': _('Saturation'),
                'subtitle': _('Super-saturate or desaturate pages'),
                'min': 0,
                'max': 400,
                'step': 1,
                'default': 100,
            },
        }

        for name, data in filters.items():
            erow = Adw.ExpanderRow()
            erow.set_title(data['title'])
            erow.set_subtitle(data['subtitle'])
            erow.set_enable_expansion(self.settings.page_filters.get(f'{name}-state'))
            erow.set_show_enable_switch(True)
            erow.connect('notify::enable-expansion', self.toggle_page_filter_state, name)

            row = Adw.SpinRow.new_with_range(data['min'], data['max'], data['step'])
            row.set_title(data['title'])
            row.set_value(self.settings.page_filters.get(name, data['default']))
            row.connect('notify::value', self.on_page_filter_changed, name)

            erow.add_row(row)
            self.page_filters_group.add(erow)

        self.set_pages_filters()

    def on_background_color_changed(self, _row, _gparam):
        value = self.background_color_row.get_selected_item().key
        if value == self.reader.background_color:
            return

        self.reader.manga.update({
            'background_color': value if value != self.settings.background_color else None,
        })

        self.set_background_color()

    def on_borders_crop_changed(self, _row, _gparam):
        value = self.borders_crop_switch.props.active
        if value == self.reader.borders_crop:
            return

        self.reader.manga.update({
            'borders_crop': value if value != self.settings.borders_crop else None,
        })

        self.reader.pager.crop_pages_borders()

    def on_borders_crop_threshold_changed(self, _row, _gparam):
        value = int(self.borders_crop_threshold_row.get_selected_item().key)
        if value == self.reader.borders_crop_threshold:
            return

        self.reader.manga.update({
            'borders_crop_threshold': value if value != self.settings.borders_crop_threshold else None,
        })

        for page in self.reader.pager.pages:
            if page.image and page.error is None:
                page.image.crop_bbox = None
                page.image.textures_crop = None
                page.image.crop_threshold = value

    def on_landscape_zoom_changed(self, _row, _gparam):
        value = self.landscape_zoom_switch.props.active
        if value == self.reader.landscape_zoom:
            return

        self.reader.manga.update({
            'landscape_zoom': value if value != self.settings.landscape_zoom else None,
        })

        self.reader.pager.rescale_pages()

    def on_page_filter_changed(self, row, _gparam, name):
        filters = self.settings.page_filters
        filters[name] = row.get_value()
        self.settings.page_filters = filters

        self.set_pages_filters()

    def on_page_numbering_changed(self, _row, _gparam):
        value = not self.page_numbering_switch.props.active
        if value == self.reader.page_numbering:
            return

        self.reader.manga.update({
            'page_numbering': value if value != self.settings.page_numbering else None,
        })

        if value and self.reader.page_numbering_defined and not self.reader.controls.is_visible:
            self.reader.page_numbering_label.set_visible(True)
        else:
            self.reader.page_numbering_label.set_visible(False)

    def on_ocr_lang_changed(self, _row, _gparam):
        value = self.ocr_lang_row.get_selected_item().key
        if value == self.reader.ocr_translator.ocr_lang:
            return

        self.reader.manga.update({
            'ocr_lang': value if value != self.settings.ocr_lang else None,
        })

    def on_scaling_changed(self, _row, _gparam):
        value = self.scaling_row.get_selected_item().key
        if value == self.reader.scaling:
            return

        self.reader.manga.update({
            'scaling': value if value != self.settings.scaling else None,
        })

        # Landscape pages zoom is enabled in RTL/LTR/Vertical reading modes only and when scaling is 'screen'
        self.landscape_zoom_switch.set_sensitive(self.reader.reading_mode != 'webtoon' and self.reader.scaling == 'screen')

        self.reader.pager.rescale_pages()

    def on_scaling_filter_changed(self, _row, _gparam):
        value = self.scaling_filter_row.get_selected_item().key
        if value == self.reader.scaling_filter:
            return

        self.reader.manga.update({
            'scaling_filter': value if value != self.settings.scaling_filter else None,
        })

        self.reader.pager.rescale_pages()

    def set_background_color(self):
        if self.reader.background_color == 'white':
            self.reader.pager.set_css_classes(['background-white'])
        elif self.reader.background_color == 'gray':
            self.reader.pager.set_css_classes(['background-gray'])
        elif self.reader.background_color == 'black':
            self.reader.pager.set_css_classes(['background-black'])
        else:
            # System style
            self.reader.pager.set_css_classes([])

    def set_pages_filters(self):
        funcs = []
        for name, value in self.settings.page_filters.items():
            if name.endswith('-state'):
                # Ignore state key
                continue
            if not self.settings.page_filters.get(f'{name}-state'):
                # Filter is off
                continue

            funcs.append(f'{name}({value}%)')

        if funcs:
            self.css_provider.load_from_string(f'.page-filters {{filter: {" ".join(funcs)};}}')  # noqa
            self.reader.overlay.add_css_class('page-filters')
        else:
            self.css_provider.load_from_string('')
            self.reader.overlay.remove_css_class('page-filters')

    def show(self):
        self.init_custom()
        self.present(self.reader.window)

    def toggle_page_filter_state(self, row, _gparam, name):
        filters = self.settings.page_filters
        filters[f'{name}-state'] = row.get_enable_expansion()
        self.settings.page_filters = filters

        self.set_pages_filters()



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

# Supported servers:
# Lelscan-VF [FR]
# FleksyScans [EN] (disabled)
# Scylla Scans [EN] (disabled)

from urllib.parse import parse_qs
from urllib.parse import urlparse

from bs4 import BeautifulSoup
import requests

from komikku.consts import USER_AGENT
from komikku.servers import Server
from komikku.servers.utils import convert_date_string
from komikku.servers.utils import get_soup_element_inner_text
from komikku.utils import get_buffer_mime_type
from komikku.utils import is_number

# https://github.com/jhin1m/fuzzy-doodle


class FuzzyDoodle(Server):
    base_url: str = None
    search_url: str = None
    manga_url: str = None
    chapter_url: str = None

    # Selectors
    imgs_src_attrs: list = ['src', 'data-src']

    search_results_selector: str = None
    search_link_selector: str = None
    search_cover_img_selector: str = None

    most_popular_results_selector: str = None
    most_popular_link_selector: str = None
    most_popular_cover_img_selector: str = None

    latest_updates_results_selector: str = None
    latest_updates_link_selector: str = None
    latest_updates_cover_img_selector: str = None

    details_name_selector: str = None
    details_cover_selector: str = None
    details_status_selector: str = None
    details_author_selector: str = None
    details_artist_selector: str = None
    details_type_selector: str = None
    details_genres_selector: str = None
    details_synopsis_selector: str = None

    chapters_selector: str = None
    chapters_title_selector: str = None
    chapters_date_selector: str = None

    chapter_pages_selector: str = None

    def __init__(self):
        if self.session is None:
            self.session = requests.Session()
            self.session.headers = {
                'User-Agent': USER_AGENT,
            }

    def get_img_src(self, element):
        for attr in self.imgs_src_attrs:
            if src := element.get(attr):
                return src

        return None

    def get_manga_data(self, initial_data):
        """
        Returns manga data by scraping manga HTML page content

        Initial data should contain at least manga's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Slug is missing in initial data'

        r = self.session_get(self.manga_url.format(initial_data['slug'], 1))
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        data = initial_data.copy()
        data.update(dict(
            authors=[],
            scanlators=[],  # not available
            genres=[],
            status=None,
            synopsis=None,
            chapters=[],
            server_id=self.id,
            cover=None,
        ))

        soup = BeautifulSoup(r.text, 'lxml')

        data['name'] = get_soup_element_inner_text(soup.select_one(self.details_name_selector), recursive=False)
        data['cover'] = self.get_img_src(soup.select_one(self.details_cover_selector))

        # Status
        status = soup.select_one(self.details_status_selector).text.strip()
        if status in ('Completed', 'Terminé'):
            data['status'] = 'complete'
        elif status in ('Ongoing', 'En cours'):
            data['status'] = 'ongoing'
        elif status in ('On-hold', ):
            data['status'] = 'suspended'
        elif status in ('Cancelled', ):
            data['status'] = 'suspended'
        elif status in ('Dropped', ):
            data['status'] = 'suspended'
        elif status in ('Hiatus'):
            data['status'] = 'hiatus'

        # Authors
        if element := soup.select_one(self.details_author_selector):
            data['authors'].append(element.text.strip())
        if element := soup.select_one(self.details_artist_selector):
            artist = element.text.strip()
            if artist not in data['authors']:
                data['authors'].append(artist)

        # Genres
        if a_elements := soup.select(self.details_genres_selector):
            for a_element in a_elements:
                data['genres'].append(a_element.text.strip())

        if element := soup.select_one(self.details_type_selector):
            type = element.text.strip()
            if type not in data['genres']:
                data['genres'].append(type)

        # Synopsis
        if elements := soup.select(self.details_synopsis_selector):
            synopsis = []
            for element in elements:
                chunk = element.text.strip()
                if chunk:
                    synopsis.append(chunk)
            data['synopsis'] = '\n\n'.join(synopsis)

        # Chapters
        data['chapters'] = self.get_manga_chapters_data(data['slug'], soup=soup)

        return data

    def get_manga_chapters_data(self, slug, num=None, soup=None, chapters=None):
        if chapters is None:
            chapters = []

        if soup is None and num is not None:
            r = self.session_get(self.manga_url.format(slug), params=dict(page=num))
            if r.status_code != 200:
                return None

            mime_type = get_buffer_mime_type(r.content)
            if mime_type != 'text/html':
                return None

            soup = BeautifulSoup(r.text, 'lxml')

        for a_element in soup.select(self.chapters_selector):
            date = a_element.select_one(self.chapters_date_selector).text.strip()
            slug = a_element.get('href').split('/')[-1]
            num = slug if is_number(slug) else None  # slug theoretically is the chapter number

            chapters.append(dict(
                slug=slug,
                title=a_element.select_one(self.chapters_title_selector).text.strip(),
                num=num,
                date=convert_date_string(date),
            ))

        if next_element := soup.select_one('#chapters-list nav > ul.pagination > li:last-child'):
            if next_url := next_element.get('onclick'):
                next_url = next_url[22:-1]
                next_num = parse_qs(urlparse(next_url).query)['page'][0]
                self.get_manga_chapters_data(slug, num=next_num, chapters=chapters)

        return list(reversed(chapters))

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data

        Currently, only pages are expected.
        """
        r = self.session_get(self.chapter_url.format(manga_slug, chapter_slug))
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = dict(
            pages=[],
        )
        for index, img_element in enumerate(soup.select(self.chapter_pages_selector)):
            data['pages'].append(dict(
                slug=None,
                image=self.get_img_src(img_element),
                index=index + 1,
            ))

        return data

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        r = self.session_get(
            page['image'],
            headers={
                'Referer': self.chapter_url.format(manga_slug, chapter_slug),
            }
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return dict(
            buffer=r.content,
            mime_type=mime_type,
            name=f'{page["index"]:04d}.{mime_type.split("/")[-1]}',  # noqa: E231
        )

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return self.manga_url.format(slug)

    def get_latest_updates(self):
        """
        Returns daily updates
        """
        r = self.session.get(self.base_url)
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        results = []
        for element in soup.select(self.latest_updates_results_selector):
            a_element = element.select_one(self.latest_updates_link_selector)
            name = element.select_one(self.latest_updates_cover_img_selector).get('alt').strip()
            img_element = element.select_one(self.latest_updates_cover_img_selector)
            last_chapter_element = element.select_one(self.latest_updates_last_chapter_selector)

            results.append(dict(
                name=name,
                slug=a_element.get('href').split('/')[-1],
                cover=self.get_img_src(img_element),
                last_chapter='#' + last_chapter_element.text.strip(),
            ))

        return results

    def get_most_populars(self):
        """
        Returns popular manga
        """
        r = self.session.get(self.base_url)
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        results = []
        for element in soup.select(self.most_popular_results_selector):
            a_element = element.select_one(self.most_popular_link_selector)
            name = element.select_one(self.most_popular_cover_img_selector).get('alt').strip()
            img_element = element.select_one(self.most_popular_cover_img_selector)

            results.append(dict(
                name=name,
                slug=a_element.get('href').split('/')[-1],
                cover=self.get_img_src(img_element),
            ))

        return results

    def search(self, term):
        r = self.session.get(
            self.search_url,
            params=dict(
                title=term,
            ),
            headers={
                'Referer': f'{self.base_url}/',
            }
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        results = []
        for element in soup.select(self.search_results_selector):
            a_element = element.select_one(self.search_link_selector)
            name = element.select_one(self.search_cover_img_selector).get('alt').strip()
            img_element = element.select_one(self.search_cover_img_selector)

            results.append(dict(
                name=name,
                slug=a_element.get('href').split('/')[-1],
                cover=self.get_img_src(img_element),
            ))

        return results
